"""
To be complete, I need:
    - PSD
    - FFT
    - Auto Corelation (on detected spikes?)

See:
    https://stackoverflow.com/questions/59265603/how-to-find-period-of-signal-autocorrelation-vs-fast-fourier-transform-vs-power
"""

import sys, math, time
from math import exp
from functools import partial
from typing import Union, Dict, List, Tuple, Optional, Optional

import numpy as np
from scipy.signal import lfilter, freqz
import scipy.signal

from PyQt5 import QtCore, QtWidgets, QtGui
import matplotlib as mpl
import matplotlib.pyplot as plt

import sanpy
from sanpy.interface.plugins import sanpyPlugin

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

def butter_sos(cutOff, fs, order=50, passType='lowpass'):
    """
    passType (str): ('lowpass', 'highpass')
    """

    '''
    nyq = fs * 0.5
    normal_cutoff = cutOff / nyq
    '''

    logger.info(f'Making butter file with order:{order} cutOff:{cutOff} passType:{passType} fs:{fs}')

    #sos = butter(order, normal_cutoff, btype='low', analog=False, output='sos')
    #sos = scipy.signal.butter(N=order, Wn=normal_cutoff, btype=passType, analog=False, fs=fs, output='sos')
    sos = scipy.signal.butter(N=order, Wn=cutOff, btype=passType, analog=False, fs=fs, output='sos')
    return sos

def spikeDetect(t, dataFiltered, dataThreshold2, startPoint=None, stopPoint=None):
    verbose = False
    #dataThreshold2 = -59.81

    g_backupFraction = 0.3  # 0.5 or 0.3

    logger.info(f'dataThreshold2:{dataThreshold2} startPoint:{startPoint} stopPoint:{stopPoint}')

    #
    # detect spikes, dataThreshold2 changes as we zoom
    Is = np.where(dataFiltered>dataThreshold2)[0]
    Is=np.concatenate(([0],Is))
    Ds=Is[:-1]-Is[1:]+1
    spikePoints = Is[np.where(Ds)[0]+1]
    spikeSeconds = t[spikePoints]

    #
    # Throw out spikes that are within refractory (ms) of previous
    spikePoints = _throwOutRefractory(spikePoints, refractory_ms=100)
    spikeSeconds = t[spikePoints]

    #
    # remove spikes outside of start/stop
    goodSpikes = []
    if startPoint is not None and stopPoint is not None:
        if verbose:
            print(f'  Removing spikes outside point range {startPoint} ... {stopPoint}')
        #print(f'    {t[startPoint]} ... {t[stopPoint]} (s)')
        for idx, spikePoint in enumerate(spikePoints):
            if spikePoint<startPoint or spikePoint>stopPoint:
                pass
            else:
                goodSpikes.append(spikePoint)
                if verbose:
                    print(f'    appending spike {idx} {t[spikePoint]}(s)')
    #
    spikePoints = goodSpikes
    spikeSeconds = t[spikePoints]

    #
    # only accept rising phase
    windowPoints = 40 # assuming 10 kHz -->> 4 ms
    goodSpikes = []
    for idx, spikePoint in enumerate(spikePoints):
        preClip = dataFiltered[spikePoint-windowPoints:spikePoint-1]
        postClip = dataFiltered[spikePoint+1:spikePoint+windowPoints]
        preMean = np.nanmean(preClip)
        postMean = np.nanmean(postClip)
        if preMean < postMean:
            goodSpikes.append(spikePoint)
        else:
            pass
            #print(f'  rejected spike {idx}, at point {spikePoint}')
    #
    spikePoints = goodSpikes
    spikeSeconds = t[spikePoints]

    #
    # throw out above take-off-potential of APs (-30 mV)
    apTakeOff = -30
    goodSpikes = []
    for idx, spikePoint in enumerate(spikePoints):
        if dataFiltered[spikePoint] < apTakeOff:
            goodSpikes.append(spikePoint)
    #
    spikePoints = goodSpikes
    spikeSeconds = t[spikePoints]

    #
    # find peak for eack spike
    # at 500 pnts, assuming spikes are slower than 20 Hz
    peakWindow_pnts = 500
    peakPoints = []
    peakVals = []
    for idx, spikePoint in enumerate(spikePoints):
        peakPnt = np.argmax(dataFiltered[spikePoint:spikePoint+peakWindow_pnts])
        peakPnt += spikePoint
        peakVal = np.max(dataFiltered[spikePoint:spikePoint+peakWindow_pnts])
        #
        peakPoints.append(peakPnt)
        peakVals.append(peakVal)

    #
    # backup each spike until pre/post is not changing
    # uses g_backupFraction to get percent of dv/dt at spike veruus each backup window
    backupWindow_pnts = 30
    maxSteps = 30  # 30 steps at 20 pnts (2ms) per step gives 60 ms
    backupSpikes = []
    for idx, spikePoint in enumerate(spikePoints):
        #print(f'=== backup spike:{idx} at {t[spikePoint]}(s)')
        foundBackupPoint = None
        for step in range(maxSteps):
            tmpPnt = spikePoint - (step*backupWindow_pnts)
            tmpSeconds = t[tmpPnt]
            preClip = dataFiltered[tmpPnt-1-backupWindow_pnts:tmpPnt-1]  # reversed
            postClip = dataFiltered[tmpPnt+1:tmpPnt+1+backupWindow_pnts]
            preMean = np.nanmean(preClip)
            postMean = np.nanmean(postClip)
            diffMean = postMean - preMean
            if step == 0:
                initialDiff = diffMean
            #print(f'  spike {idx} step:{step} tmpSeconds:{round(tmpSeconds,4)} initialDiff:{round(initialDiff,4)} diff:{round(diffMean,3)}')
            # if diffMean is 1/2 initial AP slope then accept
            if diffMean < (initialDiff * g_backupFraction):
                # stop
                if foundBackupPoint is None:
                    foundBackupPoint = tmpPnt
                #break
        #
        if foundBackupPoint is not None:
            backupSpikes.append(foundBackupPoint)
        else:
            # needed to keep spike parity
            logger.warning(f'Did not find backupSpike for spike {idx} at {t[spikePoint]}(s)')
            backupSpikes.append(spikePoint)

    #
    # use backupSpikes (points) to get each spike amplitude
    spikeAmps = []
    for idx, backupSpike in enumerate(backupSpikes):
        if backupSpikes==np.nan or math.isnan(backupSpike):
            continue
        #print('backupSpike:', backupSpike)
        footVal = dataFiltered[backupSpike]
        peakVal = peakVals[idx]
        spikeAmp = peakVal - footVal
        spikeAmps.append(spikeAmp)

    # TODO: use foot and peak to get real half/width
    theseWidths=[10,20,50,80,90]
    window_ms=100
    spikeDictList = _getHalfWidths(t, dataFiltered,
                        backupSpikes, peakPoints,
                        theseWidths=theseWidths, window_ms=window_ms)
    #for idx, spikeDict in enumerate(spikeDictList):
    #    print(idx, spikeDict)

    #
    # get estimate of duration
    # for each spike, find next downward crossing (starting at peak)
    '''
    minDurationPoints = 10
    windowPoints = 1000  # 100 ms
    fallingPoints = []
    for idx, spikePoint in enumerate(spikePoints):
        peakPoint = peakPoints[idx]
        backupPoint = backupSpikes[idx]
        if backupPoint == np.nan or math.isnan(backupPoint):
            continue
        spikeSecond = spikeSeconds[idx]
        # Not informative because we are getting spikes from absolute threshold
        thisThreshold = dataFiltered[backupPoint]
        thisPeak = dataFiltered[peakPoint]
        halfHeight = thisThreshold + (thisPeak - thisThreshold) / 2
        startPoint = peakPoint #+ 50
        postClip = dataFiltered[startPoint:startPoint+windowPoints]
        tmpFallingPoints = np.where(postClip<halfHeight)[0]
        if len(tmpFallingPoints) > 0:
            fallingPoint = startPoint + tmpFallingPoints[0]
            # TODO: check if falling point is AFTER next spike
            duration = fallingPoint - spikePoint
            if duration > minDurationPoints:
                fallingPoints.append(fallingPoint)
            else:
                print(f'  reject spike {idx} at {spikeSecond} (s), duration is {duration} points, minDurationPoints:{minDurationPoints}')
                #fallingPoints.append(fallingPoint)
        else:
            print(f'    did not find falling pnt for spike {idx} at {spikeSecond}(s) point {spikePoint}, assume it is longer than windowPoints')
            pass
            #fallingPoints.append(np.nan)
    '''

    # TODO: Package all results into a dictionary
    spikeDictList = [{}] * len(spikePoints)
    for idx, spikePoint in enumerate(spikePoints):
        '''
        spikeSecond = spikeSeconds[idx]
        peakPoint = peakPoints[idx]
        peakVal = peakVals[idx]
        fallingPoint = fallingPoints[idx]  # get rid of this
        backupSpike = backupSpikes[idx]  # get rid of this
        spikeAmp = spikeAmps[idx]  # get rid of this
        '''
        spikeDictList[idx]['spikePoint'] = spikePoint
        spikeDictList[idx]['spikeSecond'] = spikeSeconds[idx]
        spikeDictList[idx]['peakPoint'] = peakPoints[idx]
        spikeDictList[idx]['peakVal'] = peakVals[idx]
        #spikeDictList[idx]['fallingPoint'] = fallingPoints[idx]
        spikeDictList[idx]['backupSpike'] = backupSpikes[idx]
        spikeDictList[idx]['spikeAmp'] = spikeAmps[idx]
    #
    spikeSeconds = t[spikePoints]
    #return spikeSeconds, spikePoints, peakPoints, peakVals, fallingPoints, backupSpikes, spikeAmps
    # removved fallingPoints
    return spikeDictList, spikeSeconds, spikePoints, peakPoints, peakVals, backupSpikes, spikeAmps

def _getHalfWidths(t, v, spikePoints, peakPoints, theseWidths=[10,20,50,80,90], window_ms=50):
    """Get half widths.

    Args:
        t (ndarray): Time
        v (ndaray): Recording (usually curent clamp Vm)
        spikePoints (list of int): List of spike threshold crossings.
                            Usually the back-up version
        peakPoints (list of int):
        theseWidths (list of int): Specifies full-width-half maximal to calculate
        window_ms (int): Number of ms to look after the peak for downward 1/2 height crossing

    Returns:
        List of dict, one elemennt per spike
    """
    logger.info(f'theseWidths:{theseWidths} window_ms:{window_ms} spikePoints:{len(spikePoints)} peakPoints:{len(peakPoints)}')

    pointsPerMs = 10  # ToDo: generalize this
    window_pnts = window_ms * pointsPerMs

    spikeDictList = [{}] * len(spikePoints) # an empy list of dict with proper size
    for idx, spikePoint in enumerate(spikePoints):
        #print(f'Spike {idx} pnt:{spikePoint}')

        # each spike has a list of dict for each width result
        spikeDictList[idx]['widths'] = []

        # check each pre/post pnt is before next spike
        nextSpikePnt = None
        if idx < len(spikePoints)-1:
            nextSpikePnt = spikePoints[idx+1]

        peakPoint = peakPoints[idx]
        try:
            vSpike = v[spikePoint]
        except (IndexError) as e:
            logger.error(f'spikePoint:{spikePoint} {type(spikePoint)}')
        vPeak = v[peakPoint]
        height = vPeak-vSpike
        preStartPnt = spikePoint+1
        preClip = v[preStartPnt:peakPoint-1]
        postStartPnt = peakPoint+1
        postClip = v[postStartPnt:peakPoint+window_pnts]
        for width in theseWidths:
            thisHeight = vSpike + (height / width)  # search for a percent of height
            prePnt = np.where(preClip>=thisHeight)[0]
            if len(prePnt) > 0:
                prePnt = preStartPnt + prePnt[0]
            else:
                #print(f'  Error: Spike {idx} "prePnt" width:{width} vSpike:{vSpike} height:{height} thisHeight:{thisHeight}')
                prePnt = None
            postPnt = np.where(postClip<thisHeight)[0]
            if len(postPnt) > 0:
                postPnt = postStartPnt + postPnt[0]
            else:
                #print(f'  Error: Spike {idx} "postPnt" width:{width} vSpike:{vSpike} height:{height} thisHeight:{thisHeight}')
                postPnt = None
            widthMs = None
            if prePnt is not None and postPnt is not None:
                widthPnts = postPnt - prePnt
                widthMs = widthPnts / pointsPerMs
                #print(f'  width:{width} tPre:{t[prePnt]} tPost:{t[postPnt]} widthMs:{widthMs}')
                if nextSpikePnt is not None and prePnt >= nextSpikePnt:
                    print(f'  Error: Spike {idx} widthMs:{widthMs} prePnt:{prePnt} is after nextSpikePnt:{nextSpikePnt}')
                if nextSpikePnt is not None and postPnt >= nextSpikePnt:
                    print(f'  Error: Spike {idx} widthMs:{widthMs} postPnt:{postPnt} is after nextSpikePnt:{nextSpikePnt}')

            # put into dict
            widthDict = {
                'halfHeight': width,
                'risingPnt': prePnt,
                #'risingVal': defaultVal,
                'fallingPnt': postPnt,
                #'fallingVal': defaultVal,
                #'widthPnts': None,
                'widthMs': widthMs
            }
            spikeDictList[idx]['widths_' + str(width)] = widthMs
            spikeDictList[idx]['widths'].append(widthDict)
    #
    return spikeDictList

def _throwOutRefractory(spikePoints, refractory_ms=100):
    """
    spikePoints: spike times to consider
    refractory_ms:
    """
    dataPointsPerMs = 10

    before = len(spikePoints)

    # if there are doubles, throw-out the second one
    #refractory_ms = 20 #10 # remove spike [i] if it occurs within refractory_ms of spike [i-1]
    lastGood = 0 # first spike [0] will always be good, there is no spike [i-1]
    for i in range(len(spikePoints)):
        if i==0:
            # first spike is always good
            continue
        dPoints = spikePoints[i] - spikePoints[lastGood]
        if dPoints < dataPointsPerMs*refractory_ms:
            # remove spike time [i]
            spikePoints[i] = 0
        else:
            # spike time [i] was good
            lastGood = i
    # regenerate spikeTimes0 by throwing out any spike time that does not pass 'if spikeTime'
    # spikeTimes[i] that were set to 0 above (they were too close to the previous spike)
    # will not pass 'if spikeTime', as 'if 0' evaluates to False
    #if goodSpikeErrors is not None:
    #    goodSpikeErrors = [goodSpikeErrors[idx] for idx, spikeTime in enumerate(spikeTimes0) if spikeTime]
    spikePoints = [spikePoint for spikePoint in spikePoints if spikePoint]

    # TODO: put back in and log if detection ['verbose']
    after = len(spikePoints)
    logger.info(f'From {before} to {after} spikes with refractory_ms:{refractory_ms}')

    return spikePoints

def getKernel(type='sumExp', amp=5, tau1=30, tau2=70):
    """Get a kernel for convolution with a spike train."""
    N = 500  # pnts
    t = [x for x in range(N)]
    y = t

    if type == 'sumExp':
        for i in t:
            y[i] = -amp * ( exp(-t[i]/tau1) - (exp(-t[i]/tau2)) )
    #
    return y

def getSpikeTrain(numSeconds=1, fs=10000, spikeFreq=3, amp=10, noiseAmp=10):
    """Get a spike train at given frequency.

    Arguments:
        numSeconds (int): Total number of seconds
        fs (int): Sampling frequency, 10000 for 10 kH
        spikeFreq (int): Frequency of events in spike train, e.g. simulated EPSPs
        amp (float): Amplitude of sum exponential kernel (getKernel)
    """
    n = int(numSeconds * fs) # total number of samples
    numSpikes = int(numSeconds * spikeFreq)
    spikeTrain = np.zeros(n)
    start = fs / spikeFreq
    spikeTimes = np.linspace(start, n, numSpikes, endpoint=False)
    for idx, spike in enumerate(spikeTimes):
        #print(idx, spike)
        spike = int(spike)
        spikeTrain[spike] = 1

    expKernel = getKernel(amp=amp)
    epspTrain = scipy.signal.convolve(spikeTrain, expKernel, mode='same')
    # shift to -60 mV
    epspTrain -= 60

    # add noise
    if noiseAmp == 0:
        pass
    else:
        #noise_power = 0.001 * fs / noiseAmp
        #epspTrain += np.random.normal(scale=np.sqrt(noise_power), size=epspTrain.shape)
        epspTrain += np.random.normal(scale=noiseAmp, size=epspTrain.shape)

    #
    t = np.linspace(0, numSeconds, n, endpoint=True)
    #
    return t, spikeTrain, epspTrain

class fftPlugin(sanpyPlugin):
    myHumanName = 'FFT'

    #def __init__(self, myAnalysisDir=None, **kwargs):
    def __init__(self, plotRawAxis=False, ba=None, **kwargs):
        """
        Args:
            ba (bAnalysis): Not required
        """
        super(fftPlugin, self).__init__(ba=ba, **kwargs)

        self._isInited = False

        self.plotRawAxis = plotRawAxis

        self.doButterFilter = True
        self.butterOrder = 70
        self.butterCutoff = [0.7, 10]  # [low freq, high freq] for bandpass

        self._resultsDictList = []
        self._resultStr = ''

        self.isModel = False

        # only defined when running without SanPy app
        #self._analysisDir = myAnalysisDir

        self._store_ba = None  # allow switching between model and self.ba

        if self.ba is not None:
            self.fs = self.ba.fileLoader.recordingFrequency * 1000
        else:
            self.fs = None
        self.psdWindowStr = 'Hanning'  # mpl.mlab.window_hanning

        # fft Freq (Hz) resolution is fs/nfft --> resolution 0.2 Hz = 10000/50000
        self.nfft = 50000 #512 * 100

        self.signalHz = None  # assign when using fake data
        self.maxPlotHz = 10  # limit x-axis frequenccy

        #self.sos = None

        self.medianFilterPnts = 0

        # points
        self.lastLeft = None  # 0
        self.lastRight = None  # len(self.ba.sweepX())

        #if self._analysisDir is not None:
        #    # running lpugin without sanpy
        #    self.loadData(2)  # load the 3rd file in analysis dir

        self._buildInterface()

        #self._getPsd()

        self.dataLine = None
        self.dataFilteredLine = None
        self.spikesLine = None
        self.peaksLine = None
        self.dataMeanLine = None
        self.thresholdLine2 = None
        self.thresholdLine3 = None


        #self.getMean()
        self.plot()  # first plot of data
        self.replot2(switchFile=True)
        #self.replotPsd()
        #self.replot_fft()

        self._isInited = True

    @property
    def ba(self):
        return self._ba

    def _buildInterface(self):
        #self.pyqtWindow()

        # main layout
        vLayout = QtWidgets.QVBoxLayout()

        self.controlLayout = QtWidgets.QHBoxLayout()
        #
        #aLabel = QtWidgets.QLabel('fft')
        #self.controlLayout.addWidget(aLabel)

        '''
        buttonName = 'Detect'
        aButton = QtWidgets.QPushButton(buttonName)
        aButton.clicked.connect(partial(self.on_button_click,buttonName))
        self.controlLayout.addWidget(aButton)
        '''

        '''
        aLabel = QtWidgets.QLabel('mV Threshold')
        self.controlLayout.addWidget(aLabel)

        self.mvThresholdSpinBox = QtWidgets.QDoubleSpinBox()
        self.mvThresholdSpinBox.setRange(-1e9, 1e9)
        self.controlLayout.addWidget(self.mvThresholdSpinBox)
        '''

        '''
        checkboxName = 'PSD'
        aCheckBox = QtWidgets.QCheckBox(checkboxName)
        aCheckBox.setChecked(True)
        aCheckBox.stateChanged.connect(partial(self.on_checkbox_clicked, checkboxName))
        self.controlLayout.addWidget(aCheckBox)
        '''

        '''
        checkboxName = 'Auto-Correlation'
        aCheckBox = QtWidgets.QCheckBox(checkboxName)
        aCheckBox.setChecked(True)
        aCheckBox.stateChanged.connect(partial(self.on_checkbox_clicked, checkboxName))
        self.controlLayout.addWidget(aCheckBox)
        '''

        buttonName = 'Replot'
        aButton = QtWidgets.QPushButton(buttonName)
        aButton.clicked.connect(partial(self.on_button_click,buttonName))
        self.controlLayout.addWidget(aButton)

        self.resultsLabel = QtWidgets.QLabel('Results: Peak Hz=??? Ampplitude=???')
        self.controlLayout.addWidget(self.resultsLabel)

        checkboxName = 'Butter Filter'
        butterCheckBox = QtWidgets.QCheckBox(checkboxName)
        butterCheckBox.setChecked(self.doButterFilter)
        butterCheckBox.stateChanged.connect(partial(self.on_checkbox_clicked, checkboxName))
        self.controlLayout.addWidget(butterCheckBox)

        aLabel = QtWidgets.QLabel('Order')
        self.controlLayout.addWidget(aLabel)
        self.butterOrderSpinBox = QtWidgets.QSpinBox()
        self.butterOrderSpinBox.setRange(0, 2**16)
        self.butterOrderSpinBox.setValue(self.butterOrder)
        self.butterOrderSpinBox.editingFinished.connect(partial(self.on_cutoff_spinbox, aLabel))
        self.controlLayout.addWidget(self.butterOrderSpinBox)

        aLabel = QtWidgets.QLabel('Low (Hz)')
        self.controlLayout.addWidget(aLabel)
        self.lowCutoffSpinBox = QtWidgets.QDoubleSpinBox()
        self.lowCutoffSpinBox.setRange(0, 2**16)
        self.lowCutoffSpinBox.setValue(self.butterCutoff[0])
        self.lowCutoffSpinBox.editingFinished.connect(partial(self.on_cutoff_spinbox, aLabel))
        self.controlLayout.addWidget(self.lowCutoffSpinBox)

        aLabel = QtWidgets.QLabel('High')
        self.controlLayout.addWidget(aLabel)
        self.highCutoffSpinBox = QtWidgets.QDoubleSpinBox()
        self.highCutoffSpinBox.setRange(0, 2**16)
        self.highCutoffSpinBox.setValue(self.butterCutoff[1])
        self.highCutoffSpinBox.editingFinished.connect(partial(self.on_cutoff_spinbox, aLabel))
        self.controlLayout.addWidget(self.highCutoffSpinBox)

        psdWindowComboBox = QtWidgets.QComboBox()
        psdWindowComboBox.addItem('Hanning')
        psdWindowComboBox.addItem('Blackman')
        psdWindowComboBox.currentTextChanged.connect(self.on_psd_window_changed)
        self.controlLayout.addWidget(psdWindowComboBox)

        #
        vLayout.addLayout(self.controlLayout) # add mpl canvas

        self.controlLayout1_5 = QtWidgets.QHBoxLayout()

        self.fsLabel = QtWidgets.QLabel(f'fs={self.fs}')
        self.controlLayout1_5.addWidget(self.fsLabel)

        aLabel = QtWidgets.QLabel('NFFT')
        self.controlLayout1_5.addWidget(aLabel)
        self.nfftSpinBox = QtWidgets.QSpinBox()
        self.nfftSpinBox.setRange(0, 2**16)
        self.nfftSpinBox.setValue(self.nfft)
        self.nfftSpinBox.editingFinished.connect(partial(self.on_cutoff_spinbox, aLabel))
        self.controlLayout1_5.addWidget(self.nfftSpinBox)

        #self.freqResLabel = QtWidgets.QLabel(f'Freq Resolution (Hz) {round(self.fs/self.nfft,3)}')
        self.freqResLabel = QtWidgets.QLabel(f'Freq Resolution (Hz) Unknown')
        self.controlLayout1_5.addWidget(self.freqResLabel)

        aLabel = QtWidgets.QLabel('Median Filter (Pnts)')
        self.controlLayout1_5.addWidget(aLabel)
        self.medianFilterPntsSpinBox = QtWidgets.QSpinBox()
        self.medianFilterPntsSpinBox.setRange(0, 2**16)
        self.medianFilterPntsSpinBox.setValue(self.medianFilterPnts)
        self.medianFilterPntsSpinBox.editingFinished.connect(partial(self.on_cutoff_spinbox, aLabel))
        self.controlLayout1_5.addWidget(self.medianFilterPntsSpinBox)

        aLabel = QtWidgets.QLabel('Max Plot (Hz)')
        self.controlLayout1_5.addWidget(aLabel)
        self.maxPlotHzSpinBox = QtWidgets.QDoubleSpinBox()
        self.maxPlotHzSpinBox.setRange(0, 2**16)
        self.maxPlotHzSpinBox.setValue(self.maxPlotHz)
        self.maxPlotHzSpinBox.editingFinished.connect(partial(self.on_cutoff_spinbox, aLabel))
        self.controlLayout1_5.addWidget(self.maxPlotHzSpinBox)

        '''
        aLabel = QtWidgets.QLabel('Order')
        self.controlLayout1_5.addWidget(aLabel)
        self.orderSpinBox = QtWidgets.QDoubleSpinBox()
        self.orderSpinBox.setRange(-1e9, 1e9)
        self.orderSpinBox.setValue(self.order)
        self.orderSpinBox.editingFinished.connect(partial(self.on_cutoff_spinbox, aLabel))
        self.controlLayout1_5.addWidget(self.orderSpinBox)
        '''

        buttonName = 'Filter Response'
        aButton = QtWidgets.QPushButton(buttonName)
        aButton.clicked.connect(partial(self.on_button_click,buttonName))
        self.controlLayout1_5.addWidget(aButton)

        '''
        buttonName = 'Rebuild Auto-Corr'
        aButton = QtWidgets.QPushButton(buttonName)
        aButton.clicked.connect(partial(self.on_button_click,buttonName))
        self.controlLayout1_5.addWidget(aButton)
        '''

        #
        vLayout.addLayout(self.controlLayout1_5) # add mpl canvas

        #
        # second row of controls (for model)
        self.controlLayout_row2 = QtWidgets.QHBoxLayout()

        checkboxName = 'Model Data'
        self.modelDataCheckBox = QtWidgets.QCheckBox(checkboxName)
        self.modelDataCheckBox.setChecked(False)
        self.modelDataCheckBox.stateChanged.connect(partial(self.on_checkbox_clicked, checkboxName))
        self.controlLayout_row2.addWidget(self.modelDataCheckBox)

        '''
        buttonName = 'Detect'
        aButton = QtWidgets.QPushButton(buttonName)
        aButton.clicked.connect(partial(self.on_button_click,buttonName))
        self.controlLayout_row2.addWidget(aButton)
        '''

        '''
        aLabel = QtWidgets.QLabel('mvThreshold')
        self.controlLayout_row2.addWidget(aLabel)

        self.mvThresholdSpinBox = QtWidgets.QDoubleSpinBox()
        self.mvThresholdSpinBox.setValue(-52)
        self.mvThresholdSpinBox.setRange(-1000, 1000)
        self.controlLayout_row2.addWidget(self.mvThresholdSpinBox)
        '''

        # numSeconds, spikeFreq, amp, noise
        aLabel = QtWidgets.QLabel('Seconds')
        self.controlLayout_row2.addWidget(aLabel)

        self.modelSecondsSpinBox = QtWidgets.QDoubleSpinBox()
        self.modelSecondsSpinBox.setValue(20)
        self.modelSecondsSpinBox.setRange(0, 1000)
        self.controlLayout_row2.addWidget(self.modelSecondsSpinBox)

        # finalize row 2
        vLayout.addLayout(self.controlLayout_row2) # add mpl canvas

        # 3rd row
        # second row of controls (for model)
        #self.controlLayout_row3 = QtWidgets.QHBoxLayout()

        aLabel = QtWidgets.QLabel('Spike Frequency')
        self.controlLayout_row2.addWidget(aLabel)

        self.modelFrequencySpinBox = QtWidgets.QDoubleSpinBox()
        self.modelFrequencySpinBox.setValue(1)
        self.modelFrequencySpinBox.setRange(0, 100)
        self.controlLayout_row2.addWidget(self.modelFrequencySpinBox)

        aLabel = QtWidgets.QLabel('Amplitude')
        self.controlLayout_row2.addWidget(aLabel)

        self.modelAmpSpinBox = QtWidgets.QDoubleSpinBox()
        self.modelAmpSpinBox.setValue(100)
        self.modelAmpSpinBox.setRange(-100, 100)
        self.controlLayout_row2.addWidget(self.modelAmpSpinBox)

        aLabel = QtWidgets.QLabel('Noise Amp')
        self.controlLayout_row2.addWidget(aLabel)

        self.modelNoiseAmpSpinBox = QtWidgets.QDoubleSpinBox()
        self.modelNoiseAmpSpinBox.setValue(50)
        self.modelNoiseAmpSpinBox.setRange(0, 1000)
        self.controlLayout_row2.addWidget(self.modelNoiseAmpSpinBox)

        # finalize row 3
        #vLayout.addLayout(self.controlLayout_row3) # add mpl canvas

        vSplitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        vLayout.addWidget(vSplitter)

        # don't use inherited, want 3x3 with rows: 1, 1, 3 subplits
        '''
        self.mplWindow2(numRow=3)  # makes self.axs[] and self.static_canvas
        #self.vmAxes = self.axs[0]
        self.rawAxes = self.axs[0]
        self.rawZoomAxes = self.axs[1]
        self.fftAxes = self.axs[2]
        #self.autoCorrAxes = self.axs[3]  # uses spike detection
        '''

        #  mirror self.mplWindow2()
        from matplotlib.backends import backend_qt5agg
        self.fig = mpl.figure.Figure(constrained_layout=True)
        self.static_canvas = backend_qt5agg.FigureCanvas(self.fig)
        # this is really tricky and annoying
        self.static_canvas.setFocusPolicy( QtCore.Qt.ClickFocus )
        self.static_canvas.setFocus()
        #self.fig.canvas.mpl_connect('key_press_event', self.keyPressEvent)

        self.mplToolbar = mpl.backends.backend_qt5agg.NavigationToolbar2QT(self.static_canvas, self.static_canvas)

        if self.plotRawAxis:
            numRow = 3
        else:
            numRow = 2

        gs = self.fig.add_gridspec(numRow, 3)
        #self.rawAxes = self.fig.addsubplot(gs[0,:])
        if self.plotRawAxis:
            self.rawAxes = self.static_canvas.figure.add_subplot(gs[0,:])
            self.rawZoomAxes = self.static_canvas.figure.add_subplot(gs[1,:])
            # 3rd row has 3 subplots
            self.psdAxes = self.static_canvas.figure.add_subplot(gs[2,0])
            self.spectrogramAxes = self.static_canvas.figure.add_subplot(gs[2,-2])
        else:
            self.rawAxes = None
            self.rawZoomAxes = self.static_canvas.figure.add_subplot(gs[0,:])
            # 3rd row has 3 subplots
            self.psdAxes = self.static_canvas.figure.add_subplot(gs[1,0])
            self.spectrogramAxes = self.static_canvas.figure.add_subplot(gs[1,-2])

        vSplitter.addWidget(self.static_canvas) # add mpl canvas
        vSplitter.addWidget(self.mplToolbar) # add mpl canvas

        # this works
        #self.mplToolbar.hide()

        # set the layout of the main window
        #self.setLayout(vLayout)
        self.getVBoxLayout().addLayout(vLayout)
        
    def on_psd_window_changed(self, item):
        """User selected window dropdown

        Args:
            item (str):
        """
        logger.info(f'item:"{item}" {type(item)}')
        self.psdWindowStr = item
        '''
        if item == 'Hanning':
            self.psdWindow = mpl.mlab.window_hanning
        if item == 'Blackman':
            self.psdWindow = np.blackman(51)
        else:
            logger.warning(f'Item not understood {item}')
        '''
        #
        self.replot2(switchFile=False)

    def setAxis(self):
        self.replot()

    def replot(self):
        logger.info('')

        #self.getMean()
        self.replot2(switchFile=True)

    def replot2(self, switchFile=False):
        logger.info(f'switchFile:{switchFile}')

        #self.replotData(switchFile=switchFile)
        #self.replotPsd()
        #self.replot_fft()
        #self.replotAutoCorr()
        self.replot_fft2()

        if switchFile:
            self._mySetWindowTitle()

        self.static_canvas.draw()
        #plt.draw()

    def replot_fft2(self):

        def myDetrend(x):
            #print('myDetrend() x:', x.shape)
            y = plt.mlab.detrend_linear(x)
            #y = plt.mlab.detrend_mean(y)
            return y

        start = time.time()

        if self.ba is None:
            return

        if self.sweepNumber == 'All':
            logger.warning(f'fft plugin can only show one sweep, received sweepNumber:{self.sweepNumber}')
            return

        #logger.info(f'using ba: {self.ba}')
        startSec, stopSec = self.getStartStop()
        if startSec is None or stopSec is None or math.isnan(startSec) or math.isnan(stopSec):
            logger.info(f'Resetting start/stop seconds to max')
            startSec = 0
            stopSec = self.ba.fileLoader.recordingDur  #self.ba.sweepX()[-1]
        logger.info(f'Using start(s):{startSec} stop(s):{stopSec}')
        self.lastLeft = round(startSec*1000*self.ba.fileLoader.dataPointsPerMs)
        self.lastRight = round(stopSec*1000*self.ba.fileLoader.dataPointsPerMs)

        leftPoint = self.lastLeft
        rightPoint = self.lastRight

        sweepY = self.getSweep('y')
        if leftPoint<0:
            leftPoint = 0
        if rightPoint > sweepY.shape[0]:
            rightPoint = sweepY.shape[0]


        y = sweepY[leftPoint:rightPoint]

        medianPnts = self.medianFilterPnts  # 50
        if medianPnts > 0:
            logger.info(f'  Median filter with pnts={medianPnts}')
            yFiltered = scipy.ndimage.median_filter(y, medianPnts)
        else:
            logger.info('  No median filter')
            yFiltered = y

        #logger.info(f'Fetching sweepX with sweepNumber: {self.sweepNumber}')

        sweepX = self.getSweep('x')

        #logger.info(f'  sweepX: {sweepX.shape}')
        #logger.info(f'  leftSec:{sweepX[leftPoint]} rightSec:{sweepX[rightPoint-1]}')
        t = sweepX[leftPoint:rightPoint]

        dt = t[1] - t[0] #0.0001
        fs = round(1/dt)  # samples per second
        nfft = self.nfft  # The number of data points used in each block for the FFT
        #print(f'  fs:{fs} nfft:{nfft}')

        if self.plotRawAxis:
            self.rawAxes.clear()
            self.rawAxes.plot(sweepX, sweepY, '-', linewidth=1)
            # draw a rectangle to show left/right time selecction
            yMin = np.nanmin(sweepY)
            yMax = np.nanmax(sweepY)
            xRect = startSec
            yRect = yMin
            wRect = stopSec-startSec
            hRect = yMax-yMin
            rect1 = mpl.patches.Rectangle((xRect,yRect), wRect, hRect, color='gray')
            self.rawAxes.add_patch(rect1)
            self.rawAxes.set_xlim([sweepX[0], sweepX[-1]])

        #
        # replot the clip +/- std we analyzed, after detrend
        yDetrend = myDetrend(yFiltered)

        #
        if self.doButterFilter:
            #self.butterCutoff = 10  # Hz
            #self.butterOrder = 70
            #self.sos = butter_sos(self.butterCutoff, self.fs, order=self.butterOrder, passType='lowpass')
            #self.butterCutoff = [0.7, 10]
            #logger.info(f'  butterOrder:{self.butterOrder} butterCutoff:{self.butterCutoff}')
            self.sos = butter_sos(self.butterCutoff, self.fs, order=self.butterOrder, passType='bandpass')
            # filtfilt should remove phase delay in time. A forward-backward digital filter using cascaded second-order sections.
            yFiltered_Butter = scipy.signal.sosfiltfilt(self.sos, yDetrend, axis=0)
            #print('yDetrend:', yDetrend.shape, np.nanmin(yDetrend), np.nanmax(yDetrend))
            #print('yFiltered_Butter:', yFiltered_Butter.shape, np.nanmin(yFiltered_Butter), np.nanmax(yFiltered_Butter))

            # we use this for remaining
            yFiltered = yFiltered_Butter

        dataMean = np.nanmean(yDetrend)
        dataStd = np.nanstd(yDetrend)
        dataThreshold2 = dataMean + (2 * dataStd)
        dataThreshold3 = dataMean + (3 * dataStd)

        self.rawZoomAxes.clear()
        self.rawZoomAxes.plot(t, yDetrend, '-', linewidth=1)
        self.rawZoomAxes.axhline(dataMean, marker='', color='r', linestyle='--', linewidth=0.5)
        self.rawZoomAxes.axhline(dataThreshold2, marker='', color='r', linestyle='--', linewidth=0.5)
        self.rawZoomAxes.axhline(dataThreshold3, marker='', color='r', linestyle='--', linewidth=0.5)
        self.rawZoomAxes.set_xlim([t[0], t[-1]])
        #
        if self.doButterFilter:
            self.rawZoomAxes.plot(t, yFiltered, '-r', linewidth=1)

        # save yDetrend, yFiltered as csv
        '''
        print('=== FOR FERNANDO')
        import pandas as pd
        print('  t:', t.shape)
        print('  yDetrend:', yDetrend.shape)
        print('  yFiltered:', yFiltered.shape)
        tmpDf = pd.DataFrame(columns=['s', 'yDetrend', 'yFiltered'])
        tmpDf['s'] = t
        tmpDf['yDetrend'] = yDetrend
        tmpDf['yFiltered'] = yFiltered
        print(tmpDf.head())
        dfFile = 'fft-20200707-0000-16sec-31sec.csv'
        print('saving dfFile:', dfFile)
        tmpDf.to_csv(dfFile, index=False)
        print('=== END')
        '''

        # we will still scale a freq plots to [0, self.maxPlotHz]
        if self.doButterFilter:
            minPlotFreq = self.butterCutoff[0]
            maxPlotFreq = self.butterCutoff[1]
        else:
            minPlotFreq = 0
            maxPlotFreq = self.maxPlotHz #15

        #
        # spectrogram
        #self.fftAxes.clear()
        nfft2 = int(nfft / 4)
        specSpectrum, specFreqs, spec_t, specIm = self.spectrogramAxes.specgram(yFiltered, NFFT=nfft2, Fs=fs, detrend=myDetrend)
        '''
        print('=== specgram')
        print('  yFiltered:', yFiltered.shape)
        print('  specSpectrum (freq, t):', specSpectrum.shape)  # (25601, 20) is (freq, t)
        print('    ', np.nanmin(specSpectrum), np.nanmax(specSpectrum))
        print('  specFreqs:', specFreqs.shape)
        print('  spec_t:', spec_t.shape)
        print('  specIm:', type(specIm))
        '''
        mask = (specFreqs >= minPlotFreq) & (specFreqs <= maxPlotFreq)
        specSpectrum = specSpectrum[mask,:]  # TRANSPOSE
        '''
        print('  2 Transpose specSpectrum (freq, t):', specSpectrum.shape)  # (25601, 20) is (freq, t)
        print('    ', np.nanmin(specSpectrum), np.nanmax(specSpectrum))
        '''

        # careful: I want to eventually scale y-axis to [0, self.maxPlotHz]
        # for now use [minPlotFreq, maxPlotFreq]
        x_min = startSec
        x_max = stopSec
        y_min = maxPlotFreq  # minPlotFreq
        y_max = minPlotFreq  # maxPlotFreq
        extent = [x_min , x_max, y_min , y_max]

        self.spectrogramAxes.clear()
        self.spectrogramAxes.imshow(specSpectrum, extent=extent, aspect='auto', cmap='viridis')
        self.spectrogramAxes.invert_yaxis()
        self.spectrogramAxes.set_xlabel('Time (s)')
        self.spectrogramAxes.set_ylabel('Freq (Hz)')

        #
        # matplotlib psd
        if self.psdWindowStr == 'Hanning':
            psdWindow = mpl.mlab.window_hanning
        elif self.psdWindowStr == 'Blackman':
            psdWindow = np.blackman(nfft)
        else:
            logger.warning(f'psdWindowStr not understood {self.psdWindowStr}')
        '''
        print('=== calling mpl psd()')
        print('  nfft:', nfft)
        print('  fs:', fs)
        print('  myDetrend:', myDetrend)
        print('  psdWindowStr:', self.psdWindowStr)
        print('  psdWindow:', psdWindow)
        '''
        # default scale_by_freq=True
        scale_by_freq = True
        Pxx, freqs = self.psdAxes.psd(yFiltered, marker='', linestyle='-', NFFT=nfft, Fs=fs,
                                    scale_by_freq=scale_by_freq, detrend=myDetrend, window=psdWindow)

        #
        # replot matplotlib psd
        pxxLog10 = 10 * np.log10(Pxx)
        mask = (freqs >= minPlotFreq) & (freqs <= maxPlotFreq)
        pxxLog10 = pxxLog10[mask] # order matters
        freqsLog10 = freqs[mask] # order matters
        self.psdAxes.clear()
        self.psdAxes.plot(freqsLog10, pxxLog10, '-', linewidth=1)
        self.psdAxes.set_xlim([0, self.maxPlotHz])  # x-axes is frequency
        self.psdAxes.grid(True)
        #self.psdAxes.set_ylabel('10*log10(Pxx)')
        self.psdAxes.set_ylabel('PSD (dB)')
        self.psdAxes.set_xlabel('Freq (Hz)')

        #
        # get peak frequency from psd, finding max peak with width
        inflection = np.diff(np.sign(np.diff(pxxLog10)))
        peaks = (inflection < 0).nonzero()[0] + 1
        # excception ValueError
        try:
            peak = peaks[pxxLog10[peaks].argmax()]
            maxFreq = round(freqsLog10[peak],3) # Gives 0.05
            maxPsd = round(pxxLog10[peak],3) # Gives 0.05
        except(ValueError) as e:
            logger.error('BAD PEAK in pxx')
            maxFreq = []
            maxPsd =[]

        # add to plot
        self.psdAxes.plot(maxFreq, maxPsd, '.r')

        maxPsd = np.nanmax(pxxLog10)
        maxPnt = np.argmax(pxxLog10)
        maxFreq = freqsLog10[maxPnt]
        self.resultsLabel.setText(f'Results: Peak Hz={round(maxFreq,3)} Amplitude={round(maxPsd,3)}')

        #
        # print results
        # (file, startSec, stopSec, max psd amp, max psd freq)
        pStart = round(startSec,3)
        pStop = round(stopSec,3)
        pMaxFreq = round(maxFreq,3)
        pMaxPsd = round(maxPsd,3)

        printStr = f'Type\tFile\tstartSec\tstopSec\tmaxFreqPsd\tmaxPsd'  #\tmaxFreqFft\tmaxFft'
        #self.appendResultsStr(printStr)
        print('=== FFT results are:')
        print(printStr)
        printStr = f'fftPlugin\t{self.ba.fileLoader.filename}\t{pStart}\t{pStop}\t{pMaxFreq}\t{pMaxPsd}'
        self.appendResultsStr(printStr, maxFreq=pMaxFreq, maxPsd=pMaxPsd, freqs=freqsLog10, psd=pxxLog10)
        print(printStr)

        #
        # plot np fft
        # see: https://www.gw-openscience.org/tutorial05/
        # blackman is supposed to correct for low freq signal ???
        '''
        doBlackman = False
        if doBlackman:
            window = np.blackman(yFiltered.size)
            windowed_yFiltered = yFiltered*window
        else:
            windowed_yFiltered = yFiltered

        ft = np.fft.rfft(windowed_yFiltered)
        # not sure to use diff b/w [1]-[0] or fs=10000 ???
        tmpFs = t[1]-t[0]
        fftFreqs = np.fft.rfftfreq(len(windowed_yFiltered), tmpFs) # Get frequency axis from the time axis
        fftMags = abs(ft) # We don't care about the phase information here

        # find max peak with width
        inflection = np.diff(np.sign(np.diff(fftMags)))
        peaks = (inflection < 0).nonzero()[0] + 1
        peak = peaks[fftMags[peaks].argmax()]
        signal_freq = round(fftFreqs[peak],3) # Gives 0.05
        signal_mag = round(fftMags[peak],3) # Gives 0.05
        #printStr = f'FFT\t{self.ba.getFileName()}\t{pStart}\t{pStop}\t{signal_freq}\t{signal_mag}'
        #print(printStr)
        printStr += f'\t{signal_freq}\t{signal_mag}'
        print(printStr)
        self.appendResultsStr(printStr)
        #print(f'  fft signal frequency is:{signal_freq} with mag:{signal_mag}')

        # strip down to min/ax x-plot of freq
        mask = (fftFreqs>=minPlotFreq) & (fftFreqs<=maxPlotFreq)
        fftMags = fftMags[mask] # order matters
        fftFreqs = fftFreqs[mask] # order matters

        fftMags = fftMags[2:-1]
        fftFreqs = fftFreqs[2:-1]

        self.fftAxes.clear()
        self.fftAxes.semilogy(fftFreqs, fftMags, '-', linewidth=1)
        self.fftAxes.set_xlim([minPlotFreq, maxPlotFreq])  # x-axes is frequency
        self.fftAxes.set_xlabel('Freq (Hz)')
        self.fftAxes.set_ylabel('FFT Mag')
        self.fftAxes.grid(True)
        self.fftAxes.semilogy(signal_freq, signal_mag, '.r')
        '''

        stop = time.time()
        #logger.info(f'Took {stop-start} seconds.')

        #
        self.static_canvas.draw()

    def appendResultsStr(self, str, maxFreq='', maxPsd='', freqs='', psd=''):
        self._resultStr += str + '\n'
        resultDict = {
            'file': self.ba.fileLoader.filename,
            'startSec': self.getStartStop()[0],
            'stopSec': self.getStartStop()[1],
            'butterFilter': self.doButterFilter,
            'butterOrder': self.butterOrder,
            'lowFreqCutoff': self.butterCutoff[0],
            'highFreqCutoff': self.butterCutoff[1],
            'maxFreq': maxFreq,
            'maxPSD': maxPsd,
            'freqs': freqs,
            'psd': psd,
        }
        self._resultsDictList.append(resultDict)

    def getResultsDictList(self):
        return self._resultsDictList

    def getResultStr(self):
        return self._resultStr

    def old_replot_fft(self):
        self.replot_fft2()
        return

    def replotFilter(self):
        """Plot frequency response of butter sos filter."""

        logger.info('')

        #self.sos = butter_sos(self.butterCutoff, self.fs, self.butterOrder, passType='lowpass')

        fs = self.fs

        if self.sos is None:
            # filter has not been created
            self.sos = butter_sos(self.butterCutoff, self.fs, order=self.butterOrder, passType='bandpass')

        w,h = scipy.signal.sosfreqz(self.sos, worN=fs*5, fs=fs)

        print('w:', w)
        print('h:', np.abs(h))

        fig = plt.figure(figsize=(3, 3))
        ax1 = fig.add_subplot(1,1,1) #

        #plt.plot(0.5*fs*w/np.pi, np.abs(H), 'b')
        db = 20*np.log10(np.maximum(np.abs(h), 1e-5))
        #y = np.abs(H)  # 10*np.log10(np.maximum(1e-10, np.abs(H)))
        ax1.plot(w, db, '-')
        ax1.axvline(self.butterCutoff[0], color='r', linestyle='--', linewidth=0.5)
        ax1.axvline(self.butterCutoff[1], color='r', linestyle='--', linewidth=0.5)
        #y_sos = signal.sosfilt(sos, x)
        ax1.set_xlim(0, self.maxPlotHz * 2)

        ax1.set_xlabel('Frequency (Hz)')
        ax1.set_ylabel('Decibles (dB)')

        plt.show()

    def old_getPsd(self):
        """Get psd from selected x-axes range.
        """
        #logger.info(f'self.lastLeft:{self.lastLeft} self.lastRight:{self.lastRight}')
        dataFiltered = self.getSweep('filteredVm')
        leftPoint = self.lastLeft
        rightPoint = self.lastRight
        y_sos = scipy.signal.sosfilt(self.sos, dataFiltered[leftPoint:rightPoint])
        self.f, self.Pxx_den = scipy.signal.periodogram(y_sos, self.fs)

    def rebuildModel(self):
        numSeconds = self.modelSecondsSpinBox.value()
        spikeFreq = self.modelFrequencySpinBox.value()
        amp = self.modelAmpSpinBox.value()
        noiseAmp = self.modelNoiseAmpSpinBox.value()
        fs = 10000  # 10 kHz
        t, spikeTrain, data = getSpikeTrain(numSeconds=numSeconds,
                                spikeFreq=spikeFreq,
                                fs=fs,
                                amp=amp,
                                noiseAmp=noiseAmp)
        # (t, data) need to be one column (for bAnalysis)
        t = t.reshape(t.shape[0],-1)
        data = data.reshape(data.shape[0],-1)

        self._startSec = 0
        self._stopSec = numSeconds

        #t = t[:,0]
        #data = data[:,0]

        print('  model t:', t.shape)
        print('  model data:', data.shape)

        modelDict = {
            'sweepX': t,
            'sweepY': data,
            'mode': 'I-Clamp'
        }
        self._ba = sanpy.bAnalysis(fromDict=modelDict)

        #self.modelDetect()

    def loadData(self, fileIdx=0, modelData=False):
        """Load from an analysis directory. Only used when no SanPyApp."""
        if modelData:
            self.isModel = True

            # store bAnalysis so we can bring it back
            self._store_ba = self._ba
            self._store_startSec = self._startSec
            self._store_stopSec = self._stopSec

            #
            # load from a model
            #numSeconds = 50
            #spikeFreq = 1
            #amp = 20
            #  TODO: ensure we separate interface from backend !!!
            self.rebuildModel()
        else:
            self.isModel = False
            if self._analysisDir is not None:
                self._ba = self._analysisDir.getAnalysis(fileIdx)

            if self._store_ba is not None:
                self._ba = self._store_ba
                self._startSec = self._store_startSec
                self._stopSec = self._store_stopSec

        # either real ba or model
        print(self.ba)

        self.fs = self.ba.fileLoader.recordingFrequency * 1000

        self.replot2(switchFile=True)

        return

    def not_used_modelDetect(self):
        #mvThreshold = self.mvThresholdSpinBox.value() # for detection
        mvThreshold = -20

        dDict = sanpy.bDetection().getDetectionDict('SA Node')
        dDict['dvdtThreshold'] = np.nan
        dDict['mvThreshold'] = mvThreshold
        dDict['onlyPeaksAbove_mV'] = None
        dDict['doBackupSpikeVm'] = False
        self._ba.spikeDetect(dDict)

    def on_cutoff_spinbox(self, name):
        """When user sets values, rebuild low-pass filter.
        """
        self.nfft = self.nfftSpinBox.value()
        self.maxPlotHz = self.maxPlotHzSpinBox.value()
        self.medianFilterPnts = self.medianFilterPntsSpinBox.value()  # int

        self.butterOrder = self.butterOrderSpinBox.value()

        lowCutoff = self.lowCutoffSpinBox.value()  # int
        highCutoff = self.highCutoffSpinBox.value()  # int
        self.butterCutoff = [lowCutoff, highCutoff]

        self.freqResLabel.setText(f'Freq Resolution (Hz) {round(self.fs/self.nfft, 3)}')

        #self.order = self.orderSpinBox.value()
        #self.sos = butter_lowpass_sos(self.cutOff, self.fs, self.order)

    def on_checkbox_clicked(self, name, value):
        #print('on_crosshair_clicked() value:', value)
        logger.info(f'name:"{name}" value:{value}')
        isOn = value==2
        if name == 'PSD':
            self.psdAxes.set_visible(isOn)
            if isOn:
                self.vmAxes.change_.geometry(2,1,1)
                self.replotPsd()
            else:
                self.vmAxes.change_geometry(1,1,1)
            self.static_canvas.draw()
        elif name == 'Auto-Correlation':
            pass
        elif name == 'Model Data':
            self.loadData(fileIdx=0, modelData=isOn)
        elif name == 'Butter Filter':
            self.doButterFilter = value==2
            self.replot2(switchFile=False)
        else:
            logger.warning(f'name:"{name}" not understood')

    def on_button_click(self, name):
        if name == 'Replot':
            if self.isModel:
                self.rebuildModel()
                #self.replot2(switchFile=False)
            #
            self.replot2(switchFile=False)
        elif name == 'Filter Response':
            # popup new window
            self.replotFilter()
        else:
            logger.warning(f'name:"{name}" not understood')
        '''
        elif name == 'Detect':
            #mvThreshold = self.mvThresholdSpinBox.value()
            #logger.info(f'{name} mvThreshold:{mvThreshold}')
            self.modelDetect()
            self.replot2()
        elif name == 'Rebuild Auto-Corr':
            self.replotAutoCorr()
        '''

    def on_xlims_change(self, event_ax):
        return

        if not self._isInited:
            return

        #slogger.info(f'event_ax:{event_ax}')
        left, right = event_ax.get_xlim()  # seconds
        logger.info(f'left:{left} right:{right}')

        # find start/stop point from seconds in 't'
        t = self.getSweep('x')
        leftPoint = np.where(t >= left)[0]
        leftPoint = leftPoint[0]
        rightPoint = np.where(t >= right)[0]
        if len(rightPoint) == 0:
            rightPoint = len(t)  # assuming we use [left:right]
        else:
            rightPoint = rightPoint[0]
        #print('leftPoint:', leftPoint, 'rightPoint:', rightPoint, 'len(t):', len(self.t))

        # keep track of last setting, if it does not change then do nothing
        if self.lastLeft==leftPoint and self.lastRight==rightPoint:
            logger.info('left/right point same doing nothng -- RETURNING')
            return
        else:
            self.lastLeft = leftPoint
            self.lastRight = rightPoint

        #
        # get threshold fromm selection
        self.getMean()

        self.replotPsd()
        self.replot_fft()
        #self.replotAutoCorr()  # slow, requires button push

    def keyPressEvent(self, event):
        logger.info(event)
        text = super().keyPressEvent(event)

        #isMpl = isinstance(event, mpl.backend_bases.KeyEvent)
        #text = event.key
        #logger.info(f'xxx mpl key: "{text}"')

        if text == 'h':
            # hide controls (leave plots)
            self.controlLayout.hide()

        if text in ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']:
            #self.loadData yyy
            fileIdx = int(text)
            self.loadData(fileIdx)
            self.replot2(switchFile=True)

        elif text == 'f':
            self.replotFilter()

    def slot_switchFile(self,
                        ba : sanpy.bAnalysis,
                        rowDict : Optional[dict] = None,
                        replot : bool = True):
        super().slot_switchFile(ba, rowDict, replot=False)

        self.fs = self.ba.fileLoader.recordingFrequency * 1000

        self.freqResLabel.setText(f'Freq Resolution (Hz) {round(self.fs/self.nfft, 3)}')

        self.replot()

def testPlugin():
    from PyQt5 import QtWidgets
    app = QtWidgets.QApplication([])

    path = '/Users/cudmore/Sites/Sanpy/data/fft'
    ad = sanpy.analysisDir(path, autoLoad=True)

    fft = fftPlugin(ad)
    #plt.show()

    sys.exit(app.exec_())

def test_fft():
    if 0:
        # abf data
        path = '/Users/cudmore/Sites/SanPy/data/fft/2020_07_07_0000.abf'
        ba = sanpy.bAnalysis(path)

        x = ba.fileLoader.sweepX
        y = ba.fileLoader.sweepY

        # 20-30 sec
        startSec = 16.968 # good signal
        stopSec = 31.313

        # reduce to get fft with N=1024 in excel
        dataPointsPerMs = ba.fileLoader.dataPointsPerMs  # 10 for 10 kHz
        startPnt = round(startSec * 1000 * dataPointsPerMs)
        stopPnt = round(stopSec * 1000 * dataPointsPerMs)

        numPnts = stopPnt-startPnt
        print(f'N={numPnts}')

        t = x[startPnt:stopPnt]
        y = y[startPnt:stopPnt]

        #y -= np.nanmean(y)

        dt = 0.0001
        fs = 1/dt
        NFFT = 512 # 2**16 #512 * 100  # The number of data points used in each block for the FFT
        medianPnts = 50  # 5 ms

        fileName = ba.fileLoader.filename

    if 1:
        # sin wave data
        # Fixing random state for reproducibility
        np.random.seed(19680801)

        durSec = 10.24  # to get 1024 points at dt=0.01 (power of two)
        dt = 0.01
        fs = 1/dt
        nfft = 512  # The number of data points used in each block for the FFT
        medianPnts = 5  #

        t = np.arange(0, durSec, dt)
        nse = np.random.randn(len(t))
        r = np.exp(-t / 0.05)

        cnse = np.convolve(nse, r) * dt
        cnse = cnse[:len(t)]

        secondFre = 7
        y = 0.1 * np.sin(2 * np.pi * t) + cnse
        y += 0.1 * np.sin(secondFre * 2 * np.pi * t) + cnse

        fileName = 'fakeSin'

    #def replot_fft():

    #
    # filter
    yFiltered = scipy.ndimage.median_filter(y, medianPnts)

    # subsample down to 1024
    '''
    from scipy.interpolate import interp1d
    t2 = interp1d(np.arange(1024), t, 'linear')
    yFiltered2 = interp1d(t, yFiltered, 'linear')
    print(f'N2={len(t2)}')
    plt.plot(t2, yFiltered2)
    '''

    # save [t,y] to csv
    '''
    import pandas as pd
    tmpDf = pd.DataFrame(columns=['t', 'y'])
    tmpDf['t'] = t
    tmpDf['y'] = y
    csvPath = fileName + '.csv'
    print('saving csv:', csvPath)
    tmpDf.to_csv(csvPath, index=False)
    '''

    '''
    cutOff = 10 #20  # 20, cutOff of filter (Hz)
    order = 50 # 40  # order of filter
    sos = butter_lowpass_sos(cutOff, fs, order)
    '''

    #yFiltered = scipy.signal.sosfilt(sos, yFiltered)

    #
    #Fs = 1/dt  # The sampling frequency (samples per time unit)
    #NFFT = 512  # The number of data points used in each block for the FFT

    # plot
    fig, (ax0, ax1, ax2) = plt.subplots(3, 1)

    ax0.plot(t, y, 'k')
    ax0.plot(t, yFiltered, 'r')

    def myDetrend(x):
        y = plt.mlab.detrend_linear(x)
        y = plt.mlab.detrend_mean(y)
        return y

    # The power spectral density 𝑃𝑥𝑥 by Welch's average periodogram method
    print('  fs:', fs, 'nfft:', nfft)
    ax1.clear()
    Pxx, freqs = ax1.psd(yFiltered, NFFT=nfft, Fs=fs, detrend=myDetrend)
    ax1.set_xlim([0, 20])
    ax1.set_ylabel('PSD (dB/Hz)')
    #ax1.callbacks.connect('xlim_changed', self.on_xlims_change)

    '''
    ax1.clear()
    ax1.plot(freqs, Pxx)
    ax1.set_xlim([0, 20])
    '''

    '''
    # recompute the ax.dataLim
    ax1.relim()
    # update ax.viewLim using the new dataLim
    ax1.autoscale_view()
    plt.draw()
    '''

    '''
    ax2.plot(freqs, Pxx)
    ax2.set_xlim([0, 20])
    '''

    maxPsd = np.nanmax(Pxx)
    maxPnt = np.argmax(Pxx)
    print(f'Max PSD freq is {freqs[maxPnt]} with power {maxPsd}')

    scipy_f, scipy_Pxx = scipy.signal.periodogram(yFiltered, fs)
    ax2.plot(scipy_f[1:-1], scipy_Pxx[1:-1])  # drop freq 0
    ax2.set_xlim([0, 20])

    ax2.set_xlabel('Frequency (Hz)')
    ax2.set_ylabel('scipy_Pxx')

    #
    plt.show()

def testFilter():
    butterOrder = 50
    butterCutoff = [0.7, 10]
    fs = 10000
    #nyq = fs * 0.5
    #normal_cutoff = butterCutoff / nyq
    #normal_cutoff = butterCutoff
    passType = 'bandpass'

    #sos = scipy.signal.butter(N=butterOrder, Wn=butterCutoff, btype=passType, analog=False, fs=fs, output='sos')
    sos = butter_sos(butterCutoff, fs, order=butterOrder, passType='bandpass')

    w,h = scipy.signal.sosfreqz(sos, worN=fs*5, fs=fs)

    db = 20*np.log10(np.maximum(np.abs(h), 1e-5))

    print('w:', w)
    print('db:', db)

    fig = plt.figure(figsize=(3, 3))
    ax1 = fig.add_subplot(1,1,1) #

    ax1.plot(w, db, '-')
    ax1.set_xlim([0, 15])

    plt.show()

def main():
    path = '/home/cudmore/Sites/SanPy/data/19114001.abf'
    ba = sanpy.bAnalysis(path)
    ba.spikeDetect()
    print(ba.numSpikes)

    import sys
    app = QtWidgets.QApplication([])
    fft = fftPlugin(ba=ba)
    fft.show()
    sys.exit(app.exec_())

if __name__ == '__main__':
    #test_fft()
    #testFilter()
    main()



#

from collections import OrderedDict

# todo: move these to a json file !!!
# A list of human readable stats and how to map them to backend
# Each key, like 'Take Off Potential (mV)' is a y-stat
statList = OrderedDict()
'''
statList['Inter-Spike-Interval (ms)'] = {
	'yStat': 'isi_ms',
	'yStatUnits': 'ms',
	'xStat': 'thresholdPnt',
	'xStatUnits': 'Points'
	}
'''
statList['Spike Frequency (Hz)'] = {
	'yStat': 'spikeFreq_hz',
	'yStatUnits': 'Hz',
	'xStat': 'thresholdPnt',
	'xStatUnits': 'Points'
	}
statList['Cycle Length (ms)'] = {
	'yStat': 'cycleLength_ms',
	'yStatUnits': 'ms',
	'xStat': 'thresholdPnt',
	'xStatUnits': 'Points'
	}
statList['Take Off Potential (mV)'] = {
	'yStat': 'thresholdVal',
	'yStatUnits': 'mV',
	'xStat': 'thresholdPnt',
	'xStatUnits': 'Points'
	}
statList['AP Peak (mV)'] = {
	'yStat': 'peakVal',
	'yStatUnits': 'mV',
	'xStat': 'peakPnt',
	'xStatUnits': 'Points'
	}
#spikeDict['peakVal'] - spikeDict['thresholdVal']
statList['AP Height (mV)'] = {
	'yStat': 'peakHeight',
	'yStatUnits': 'mV',
	'xStat': 'peakPnt',
	'xStatUnits': 'Points'
	}
statList['Pre AP Min (mV)'] = {
	'yStat': 'preMinVal',
	'yStatUnits': 'mV',
	'xStat': 'preMinPnt',
	'xStatUnits': 'Points'
	}
statList['Post AP Min (mV)'] = {
	'yStat': 'postMinVal',
	'yStatUnits': 'mV',
	'xStat': 'postMinPnt',
	'xStatUnits': 'Points'
	}
# todo: fix this
statList['Early Diastolic Depol Rate (dV/s)'] = {
	'yStat': 'earlyDiastolicDurationRate',
	'yStatUnits': 'dV/s',
	'xStat': '',
	'xStatUnits': ''
	}
# todo: fix this
statList['Early Diastolic Duration (ms)'] = {
	'yStat': 'earlyDiastolicDuration_ms',
	'yStatUnits': 'dV/s',
	'xStat': 'thresholdPnt',
	'xStatUnits': 'Points'
	}

statList['Diastolic Duration (ms)'] = {
	'yStat': 'diastolicDuration_ms',
	'yStatUnits': 'dV/s',
	'xStat': 'thresholdPnt',
	'xStatUnits': 'Points'
	}
statList['Max AP Upstroke (mV)'] = {
	'yStat': 'preSpike_dvdt_max_val',
	'yStatUnits': 'dV/s',
	'xStat': 'preSpike_dvdt_max_pnt',
	'xStatUnits': 'Points'
	}
statList['Max AP Upstroke (dV/dt)'] = {
	'yStat': 'preSpike_dvdt_max_val2',
	'yStatUnits': 'dV/dt',
	'xStat': 'preSpike_dvdt_max_pnt',
	'xStatUnits': 'Points'
	}
statList['Max AP Repolarization (mV)'] = {
	'yStat': 'postSpike_dvdt_min_val',
	'yStatUnits': 'mV',
	'xStat': 'postSpike_dvdt_min_pnt',
	'xStatUnits': 'Points'
	}
# todo: fix this
statList['AP Duration (ms)'] = {
	'yStat': 'apDuration_ms',
	'yStatUnits': 'ms',
	'xStat': 'thresholdPnt',
	'xStatUnits': 'Points'
	}

# new 20210211
statList['Half Width 10 (ms)'] = {
	'yStat': 'widths_10',
	'yStatUnits': 'ms',
	'xStat': '',
	'xStatUnits': ''
	}
statList['Half Width 20 (ms)'] = {
	'yStat': 'widths_20',
	'yStatUnits': 'ms',
	'xStat': '',
	'xStatUnits': ''
	}
statList['Half Width 50 (ms)'] = {
	'yStat': 'widths_50',
	'yStatUnits': 'ms',
	'xStat': '',
	'xStatUnits': ''
	}
statList['Half Width 80 (ms)'] = {
	'yStat': 'widths_80',
	'yStatUnits': 'ms',
	'xStat': '',
	'xStatUnits': ''
	}
statList['Half Width 90 (ms)'] = {
	'yStat': 'widths_90',
	'yStatUnits': 'ms',
	'xStat': '',
	'xStatUnits': ''
	}
# sa node specific
'''
statList['Region'] = {
	'yStat': 'Region',
	'yStatUnits': '',
	'xStat': '',
	'xStatUnits': ''
	}
'''
# kymograph analysis
statList['Ca++ Delay (s)'] = {
	'yStat': 'caDelay_sec',
	'yStatUnits': 's',
	'xStat': '',
	'xStatUnits': ''
	}
statList['Ca++ Width (ms)'] = {
	'yStat': 'caWidth_ms',
	'yStatUnits': 'ms',
	'xStat': '',
	'xStatUnits': ''
	}



from functools import partial

from PyQt5 import QtWidgets, QtGui, QtCore

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin

class ComboBox(QtWidgets.QComboBox):
    # https://code.qt.io/cgit/qt/qtbase.git/tree/src/widgets/widgets/qcombobox.cpp?h=5.15.2#n3173
    def paintEvent(self, event):
        
        painter = QtWidgets.QStylePainter(self)
        painter.setPen(self.palette().color(QtGui.QPalette.Text))

        # draw the combobox frame, focusrect and selected etc.
        opt = QtWidgets.QStyleOptionComboBox()
        self.initStyleOption(opt)
        painter.drawComplexControl(QtWidgets.QStyle.CC_ComboBox, opt)

        if self.currentIndex() < 0:
            opt.palette.setBrush(
                QtGui.QPalette.ButtonText,
                opt.palette.brush(QtGui.QPalette.ButtonText).color().lighter(),
            )
            if self.placeholderText():
                opt.currentText = self.placeholderText()

        # draw the icon and text
        painter.drawControl(QtWidgets.QStyle.CE_ComboBoxLabel, opt)

# setSpikeStatEvent = {
#     'spikeList': [],
#     'colStr': '',
#     'value': ''
# }

class SetSpikeStat(sanpyPlugin):
    """Plugin to provide an interface to set spike stats like condition, userType, isBad, etc.
    
    Get stat names and variables from sanpy.bAnalysisUtil.getStatList()
    """
    signalSetSpikeStat = QtCore.Signal(dict)

    myHumanName = 'Set Spike Stats'
    showInMenu = False
    
    def __init__(self, **kwargs):
        logger.info('')
        super().__init__(**kwargs)

        self.trueFalseItems = ['True', 'False']
        self.userTypeItems = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9']
        self._buildUI()

        self._setState()

    def _buildUI(self):
        vBoxLayout = self.getVBoxLayout()

        # first row
        hBoxLayout = QtWidgets.QHBoxLayout()
        vBoxLayout.addLayout(hBoxLayout)

        # a button to set
        self._setButton = QtWidgets.QPushButton('Set')
        self._setButton.clicked.connect(self.on_set_button)
        hBoxLayout.addWidget(self._setButton, alignment=QtCore.Qt.AlignLeft)

        # a label to report spike selection
        self._spikeLabel = QtWidgets.QLabel('0 Spikes')
        hBoxLayout.addWidget(self._spikeLabel, alignment=QtCore.Qt.AlignLeft)

        # second row
        hBoxLayout = QtWidgets.QHBoxLayout()
        vBoxLayout.addLayout(hBoxLayout)

        # popup with column types
        _items = ['condition', 'userType', 'isBad']
        self._statComboBox = ComboBox()
        for item in _items:
            self._statComboBox.addItem(item)
        # self._statComboBox.setPlaceholderText("Some placeholder text here")         
        # self._statComboBox.setCurrentIndex(-1)
        self._statComboBox.setCurrentIndex(0)
        self._statComboBox.currentIndexChanged.connect(self.on_stat_change)
        hBoxLayout.addWidget(self._statComboBox, alignment=QtCore.Qt.AlignLeft)

        # line edit for user to enter values
        self._lineEdit = QtWidgets.QLineEdit('')
        #self._lineEdit.setReadOnly(True)  # for now our 1 edit widget is not editable
        #self._lineEdit.setKeyboardTracking(False) # don't trigger signal as user edits
        #self._lineEdit.setValidator(QIntValidator())
        #self._lineEdit.setMaxLength(4)
        self._lineEdit.setAlignment(QtCore.Qt.AlignLeft)
        self._lineEdit.editingFinished.connect(partial(self.on_text_edit, self._lineEdit))
        hBoxLayout.addWidget(self._lineEdit, alignment=QtCore.Qt.AlignLeft)

        # popup with true false (for some column types like 'isBad')
        # switch to userType [1,2,3,4, ...]
        _items = ['True', 'False']
        self._valueComboBox = ComboBox()
        for item in _items:
            self._valueComboBox.addItem(item)
        # self._valueComboBox.setPlaceholderText("Some placeholder text here")         
        # self._valueComboBox.setCurrentIndex(-1)
        self._valueComboBox.setEnabled(False)
        self._valueComboBox.setCurrentIndex(0)
        hBoxLayout.addWidget(self._valueComboBox, alignment=QtCore.Qt.AlignLeft)

    def _setState(self):
        """Set all ui based on state, the number of spikes and the popups selected.
        """

        _selectedSpikes = self.getSelectedSpikes()

        if  _selectedSpikes == []:
            self._spikeLabel.setText('0 Spikes')
            self._setButton.setEnabled(False)
            self._statComboBox.setEnabled(False)
            self._lineEdit.setEnabled(False)
            self._valueComboBox.setEnabled(False)
            return

        self._setButton.setEnabled(True)
        self._statComboBox.setEnabled(True)

        self._spikeLabel.setText(f'{len(_selectedSpikes)} Spikes')

        # isBad sets combobox to trueFalseItems
        # userType sets combobox to userTypeItems
        colStr = self._statComboBox.currentText()

        # like a radio button, either one or the other but not both (disjoint)
        trueFalseEnabled = colStr in ['isBad', 'userType']
        self._lineEdit.setEnabled(not trueFalseEnabled)
        self._valueComboBox.setEnabled(trueFalseEnabled)

        # switch items in combobox
        if 'userType' in colStr:
            items = self.userTypeItems
        elif colStr == 'isBad':
            items = self.trueFalseItems
        else:
            items = None
        
        if trueFalseEnabled and items is not None:
            # set items in combo box and select item 0
            _numItems = self._valueComboBox.count()
            self._valueComboBox.clear()
            #logger.info(f'adding {items}')
            for item in items:
                self._valueComboBox.addItem(item)
            self._valueComboBox.setCurrentIndex(0)

    def on_set_button(self):
        """Respond to user clicking the 'set' button.
        """
        colStr = self._statComboBox.currentText()

        if colStr in ['isBad', 'userType']:
            value = self._valueComboBox.currentText()
        else:
            value = self._lineEdit.text()

        if colStr == 'userType':
            value = int(value)
            
        #eventDict = setSpikeStatEvent
        setSpikeStatEvent = {}
        setSpikeStatEvent['spikeList'] = self.getSelectedSpikes()
        setSpikeStatEvent['colStr'] = colStr
        setSpikeStatEvent['value'] = value

        logger.info(f'  -->> emit setSpikeStatEvent:{setSpikeStatEvent}')
        self.signalSetSpikeStat.emit(setSpikeStatEvent)
        
    def on_stat_change(self, index : int):
        """User has selected an item in the stat combo box.
        """
        self._setState()

    def on_text_edit(self, widget : QtWidgets.QLineEdit):
        """User has finished editing the text for a column.
        """
        text = widget.text()
        logger.info(f'{text}')

    def replot(self):
        """Inherited.
        """
        self._setState()

    def selectSpikeList(self):
        """Inherited.
        """
        self.replot()

def testSetSpikeStat():
    import sys
    
    # load file
    path = '/Users/cudmore/Dropbox/data/san-ap/20191009_0006.abf'
    ba = sanpy.bAnalysis(path)
 
    # get all detection presets
    _detectionParams = sanpy.bDetection()    
    # select 'SA Node' presets
    _dDict = _detectionParams.getDetectionDict('SA Node')

    ba.spikeDetect(_dDict)
    print(ba)

    # run the app
    app = QtWidgets.QApplication([])

    sss = SetSpikeStat(ba=ba)
    sss.show()

    sys.exit(app.exec_())

if __name__ == '__main__':
    testSetSpikeStat()




from .sanpyLogger import *

from ._util import *
from .analysisUtil import *

from .bAnalysis_ import bAnalysis
from .analysisDir import *
from .analysisPlot import *
from .bAnalysisUtil import *
from .atfStim import *
from .bAnalysisResults import *

from .bAbfText import bAbfText
from .bExport import bExport

from .version import analysisVersion
from .version import interfaceVersion

from .bDetection import bDetection

from .kymAnalysis import kymAnalysis
from ._util import _loadLineScanHeader

from .fileloaders import *


from pprint import pprint

from typing import List, Union, Optional  # Callable, Iterator, Optional
import warnings  # to catch np.polyfit -->> RankWarning: Polyfit may be poorly conditioned

import numpy as np
import scipy.signal

import sanpy

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

def getMedian_Filter(y : np.ndarray, windowPnts : int = 5, verbose = False):
	"""Get median filtered version of y using scipy.signal.medfilt2d
	"""
	if verbose:
		logger.info('')
	filtered = scipy.signal.medfilt2d(y, windowPnts)
	return filtered

def getSavitzkyGolay_Filter(y : np.ndarray, pnts : int = 5, poly : int = 2, verbose = False):
	"""Get SavitzkyGolay filtered version of y using scipy.signal.savgol_filter
	"""
	if verbose:
		logger.info('')
	filtered = scipy.signal.savgol_filter(
								y, pnts, poly,
								mode='nearest',
								axis=0)
	return filtered

def getFirstDerivative(y):
	"""Get the first derivative from a time-series.
	
	Args:
		y (numpy.ndarr)
	"""
	firstDerivative = np.diff(y, axis=0)
	firstDerivative = np.insert(firstDerivative, 0, np.nan)
	return firstDerivative
	
def getSecondDerivative(y):
	"""Get the second derivative from a time-series.
	
	Args:
		y (numpy.ndarray)
	"""
	firstDerivative = getFirstDerivative(y)
	secondDerivative = np.diff(firstDerivative, axis=0)
	secondDerivative = np.insert(secondDerivative, 0, np.nan)
	return secondDerivative

def getThreshold_firstDerivative(y, threshold : float, firstDerivative = None):
	"""Find threshold crossings in the first derivative.
	"""
	if firstDerivative is None:
		firstDerivative = getFirstDerivative(y)
	
	Is = np.where(firstDerivative>threshold)[0]
	Is = np.concatenate(([0],Is))
	Ds = Is[:-1] - Is[1:]+1
	thresholdPoints = Is[np.where(Ds)[0]+1]
	return thresholdPoints

def getThreshold_vm(y, threshold : float):
	"""Find threshold crossings in Vm (membrane potential).
	"""
	Is = np.where(y>threshold)[0]
	Is = np.concatenate(([0],Is))
	Ds = Is[:-1] - Is[1:]+1
	thresholdPoints = Is[np.where(Ds)[0]+1]
	return thresholdPoints

def reduceByRefractory(spikeTimes : List[int], refractoryPnts : int):
	"""If there are fast-spikes, throw-out the second one.
	
	Args:
        spikeTimes: list of spike time threshold crossing
        refractoryPnts: 
	"""

	#refractory_ms = 20 #10 # remove spike [i] if it occurs within refractory_ms of spike [i-1]

	spikeTimes = spikeTimes.copy()  # make a copy of the list, we are modifying it
	
	lastGood = 0 # first spike [0] will always be good, there is no spike [i-1]
	for i in range(len(spikeTimes)):
		if i==0:
			# first spike is always good
			continue
		dPoints = spikeTimes[i] - spikeTimes[lastGood]
		if dPoints < refractoryPnts:
			# remove spike time [i]
			spikeTimes[i] = 0
		else:
			# spike time [i] was good
			lastGood = i

	# regenerate spikeTimes0 by throwing out any spike time that does not pass 'if spikeTime'
	# spikeTimes[i] that were set to 0 above (they were too close to the previous spike)
	newSpikeTimes = [spikeTime for spikeTime in spikeTimes if spikeTime]
	return newSpikeTimes

def refineWithDerivative(y : np.ndarray, spikeTimes : List[int],
						dvdtPreWindow_pnts : int,
						firstDeriv = None,
						secondDeriv = None,
						verbose=True):
	"""Refine spike times by looking for zero crossing in first derivative.
	
	Additionally, look for peak in second derivative.

	Args:
		y (np.ndarrya) Signal to examine, usually membrane potential
		spikeTimes (list[int]): 
		firstDeriv (np.ndarry): Specify this to speed execution time
		secondDeriv (np.ndarry): Specify this to speed execution time

	TODO: pass (prePnts, postPnts) as parameter. 
	
	See: refineWithPercentOfMax
	"""
	
	if firstDeriv is None:
		firstDeriv = getFirstDerivative(y)
	if secondDeriv is None:
		secondDeriv = getSecondDerivative(y)

	n = len(spikeTimes)
	footPntList = [None] * n

	for idx,footPnt in enumerate(spikeTimes):
		lastCrossingPnt = footPnt
		# move forwared a bit in case we are already in a local minima ???
		#footPnt += 2  # TODO: add as param
		
		preStart = footPnt - dvdtPreWindow_pnts
		preClip = firstDeriv[preStart:footPnt]

		zero_crossings = np.where(np.diff(np.sign(preClip)))[0]  # find where derivative flips sign (crosses 0)
		if len(zero_crossings)==0:
			foundFoot = False
			if verbose:
				logger.error(f'  no foot for peak {idx} at pnt {footPnt} ... did not find zero crossings')
		else:
			foundFoot = True
			lastCrossingPnt = preStart + zero_crossings[-1]

		footPntList[idx] = lastCrossingPnt # was this and worked, a bit too early

		# find peak in second derivative
		if foundFoot:
			preStart2 = lastCrossingPnt
			postPnts = dvdtPreWindow_pnts
			footPnt2 = preStart2 + postPnts
			preClip2 = secondDeriv[preStart2:footPnt2]
			peakPnt2 = np.argmax(preClip2)
			peakPnt2 += preStart2
			#
			footPntList[idx] = peakPnt2

	return footPntList

def refineWithPercentOfMax(y, spikeTimes, dvdt_percentOfMax, firstDeriv=None, verbose=False):
	"""Backup spike times by looking for a percent of maximum in the first derivative.
	
	See: refineWithDerivative
	"""
	window_pnts = 30
	
	if firstDeriv is None:
		firstDeriv = getFirstDerivative(y)

	newSpikeTimes = []
	spikeErrorList = []

	for idx,spikeTime in enumerate(spikeTimes):
		preDerivClip = firstDeriv[spikeTime-window_pnts:spikeTime] # backwards
		postDerivClip = firstDeriv[spikeTime:spikeTime+window_pnts] # forwards

		if len(preDerivClip) == 0:
			logger.error('')
			print('FIX ERROR: spikeDetect_dvdt()',
					'spike', idx, 'at pnt', spikeTime,
					'window_pnts:', window_pnts,
					'len(preDerivClip)', len(preDerivClip))#preDerivClip = np.flip(preDerivClip)

		# look for % of max in dvdt
		try:
			peakPnt = np.argmax(postDerivClip)
			peakPnt += spikeTime
			peakVal = firstDeriv[peakPnt]

			percentMaxVal = peakVal * dvdt_percentOfMax # value we are looking for in dv/dt
			preDerivClip = np.flip(preDerivClip) # backwards
			tmpWhere = np.where(preDerivClip<percentMaxVal)
			tmpWhere = tmpWhere[0]  # we flipped it
			if len(tmpWhere) > 0:
				threshPnt2 = np.where(preDerivClip<percentMaxVal)[0][0]
				threshPnt2 = (spikeTime) - threshPnt2
				#threshPnt2 -= 1 # backup by 1 pnt
				newSpikeTimes.append(threshPnt2)
				spikeErrorList.append(None)

			else:
				errorType = 'dvdt Percent'
				errStr = f"Did not find dvdt_percentOfMax: percent={dvdt_percentOfMax} peak dV/dt is {round(peakVal,2)}"
				eDict = getErrorDict(idx, spikeTime, errorType, errStr) # spikeTime is in pnts
				if verbose:
					logger.error(eDict)
				spikeErrorList.append(eDict)
				# always append, do not REJECT spike if we can't find % in dv/dt
				newSpikeTimes.append(spikeTime)
		except (IndexError, ValueError) as e:
			print('   FIX ERROR: refineWithPercentOfMax() looking for dvdt_percentOfMax')
			print('	  ', 'IndexError for spike', idx, spikeTime)
			print('	  ', e)
			# always append, do not REJECT spike if we can't find % in dv/dt
			newSpikeTimes.append(spikeTime)
			spikeErrorList.append(None)

	return newSpikeTimes, spikeErrorList

def keepPeaksAbove(y, spikeTimes, onlyPeaksAbove_mV : float, verbose=False):
	"""Keep peaks that are aboe a threshold.

	Args:
		onlyPeaksAbove_mV (float): Keep peaks above this threshold (discard peaks below).
	"""
	if onlyPeaksAbove_mV is None:
		return spikeTimes
	goodSpikeTimes = [spikeTime for spikeTime in spikeTimes if y[spikeTime] > onlyPeaksAbove_mV]
	if verbose:
		logger.info('reduce peaks from {len(spikeTimes)} to len(goodSpikeTimes})')
	return goodSpikeTimes

def keepPeaksBelow(y, spikeTimes, onlyPeaksBelow_mV : float, verbose=False):
	"""Keep peaks that are below a threshold.

	Args:
		onlyPeaksBelow_mV (float): Keep peaks below this threshold (discard peaks above).
	"""
	if onlyPeaksBelow_mV is None:
		return spikeTimes
	goodSpikeTimes = [spikeTime for spikeTime in spikeTimes if y[spikeTime] < onlyPeaksBelow_mV]
	if verbose:
		logger.info('reduce peaks from {len(spikeTimes)} to len(goodSpikeTimes})')
	return goodSpikeTimes

def getPeaks(y, spikeTimes, peakWindow_pnts : int):
	"""Get the peak following the threshold crossing.

	Args:
		peakWindow_pnts (int): Window to look for peaks, after each spikeTime.
	"""
	peakPnts = []
	for spikeTime in spikeTimes:
		peakPnt = np.argmax(y[spikeTime:spikeTime+peakWindow_pnts])
		peakPnt += spikeTime
		peakPnts.append(peakPnt)
	return peakPnts

def getHalfWidth(y, spikePnts, spikePeakPnts,
				hwWindowPnts : int,
				dataPointsPerMs,
				halfHeights=[10,20,50,80,90],
				verbose=False):
	"""Get half-width for a list of spikes.
	"""
	hwList = [None] * len(spikePnts)
	hwErrorList = [None] * len(spikePnts)
	for idx, spikePnt in enumerate(spikePnts):
		peakPnt = spikePeakPnts[idx]
		hwList[idx], hwErrorList[idx] = getHalfWidth_(y, spikePnt, peakPnt, 
							idx,
							hwWindowPnts,
							dataPointsPerMs,
							halfHeights=halfHeights, verbose=verbose)

	return hwList, hwErrorList

def getHalfWidth_(vm, thresholdPnt, peakPnt,
				peakIdx : int,
				hwWindowPnts : int,
				dataPointsPerMs : Optional[int] = None,
				halfHeights=[10,20,50,80,90], verbose=False):
	"""
	Get half-widhts for one spike.

	Args:
		vm ():
		thresholdPnt (int): AP threshold crossing
		peakPnt (int): AP peak
		peakIdx (int): The peak index within all peaks. Needed to trigger a proper error
		hwWindowPnts (int): Window to look after peakPnt for falling vm
		dataPointsPerMs (int):
		halfHeightList (list): List of half-height [10,20,50,80,90]
	"""

	#halfWidthWindow_ms = None
	#if dataPointsPerMs is not None:
	#	halfWidthWindow_ms = hwWindowPnts / dataPointsPerMs

	thresholdVal = vm[thresholdPnt]
	peakVal = vm[peakPnt]
	spikeHeight = peakVal - thresholdVal

	spikeSecond = thresholdPnt / dataPointsPerMs /1000
	peakSec = peakPnt / dataPointsPerMs /1000

	widthDictList = []
	errorList = []

	# clear out any existing list
	#spikeDict[iIdx]['widths'] = []

	tmpErrorType = None
	for j, halfHeight in enumerate(halfHeights):
		# halfHeight in [20, 50, 80]

		# search rising/falling phase of vm for this vm
		thisVm = thresholdVal + spikeHeight * (halfHeight * 0.01)

		#todo: logic is broken, this get over-written in following try
		widthDict = {
			'halfHeight': halfHeight,
			'risingPnt': None,
			'fallingPnt': None,
			'widthPnts': None,
			'widthMs': float('nan')
		}
		#widthMs = float('nan')
		currentErrorType = 'no error'
		try:
			postRange = vm[peakPnt:peakPnt+hwWindowPnts]
			fallingPnt = np.where(postRange<thisVm)[0] # less than
			if len(fallingPnt)==0:
				# no falling pnts found within hwWindowPnts
				currentErrorType = 'falling point'
				raise IndexError
			fallingPnt = fallingPnt[0] # first falling point
			fallingPnt += peakPnt
			fallingVal = vm[fallingPnt]

			# use the post/falling to find pre/rising
			preRange = vm[thresholdPnt:peakPnt]
			risingPnt = np.where(preRange>fallingVal)[0] # greater than
			if len(risingPnt)==0:
				currentErrorType = 'rising point'
				raise IndexError
			risingPnt = risingPnt[0] # first rising point
			risingPnt += thresholdPnt

			# actual width (pnts, ms)
			widthPnts = fallingPnt - risingPnt
			if dataPointsPerMs is not None:
				widthMs = widthPnts / dataPointsPerMs
			else:
				widthMS = None
			# assign
			widthDict['halfHeight'] = halfHeight
			widthDict['risingPnt'] = risingPnt
			widthDict['fallingPnt'] = fallingPnt
			widthDict['widthPnts'] = widthPnts
			widthDict['widthMs'] = widthMs  # can be none

			#widthDictList.append(None)
			errorList.append(None)

		except (IndexError) as e:
			errorType = 'Spike Width'
			errorStr = (f'Half width {halfHeight} error in "{currentErrorType}" '
					f"with hwWindowPnts:{hwWindowPnts} "
					f'searching for Vm:{round(thisVm,2)} from peak sec {round(peakSec,2)}'
					)

			eDict = getErrorDict(peakIdx, thresholdPnt, errorType, errorStr) # spikeTime is in pnts
			errorList.append(eDict)
			if verbose:
				logger.error(f'peakIdx:{peakIdx} spikeSecond:{spikeSecond} halfHeight:{halfHeight} eDict:{eDict}')

		#
		#spikeDict[iIdx]['widths_'+str(halfHeight)] = widthMs
		widthDictList.append(widthDict)

	#
	return widthDictList, errorList

def getMaximumDepolaizingPotential(y, spikeTimes : List[int],
				mdp_pre_pnts : int, avgWindow_pnts : int,
				verbose=False):
	"""Get the maximum depolaizing potential between spikes.
	
	This is a very 'cardiac' measure and is often difficulat to interpret if
	cells do not have a 'ventricular' or rhythmic like AP waveform.

	Args:
		y (np.ndarray):
		spikeTimes (list[int]):
		mdp_pre_nts (int)
		avgWindow_pnts (int)
	"""

	preMinPntList = [None] * len(spikeTimes)
	errorList = [[]] * len(spikeTimes)  # list of spikes, list within each spike

	for idx, spikePnt in enumerate(spikeTimes):
		# other algorithms look between spike[i-1] and spike[i]
		# here we are looking in a predefined window
		startPnt = spikeTimes[idx]-mdp_pre_pnts

		if startPnt < 0:
			errorType = 'mdp'
			detailStr = 'Went past start of recording, setting start to 0'
			errorDict = getErrorDict(idx, spikePnt, errorType, detailStr)
			errorList[idx].append(errorDict)
			startPnt = 0

		preRange = y[startPnt:spikeTimes[idx]]
		preMinPnt = np.argmin(preRange)
		preMinPnt += startPnt
		# the pre min is actually an average around the real minima
		avgRange = y[preMinPnt-avgWindow_pnts:preMinPnt+avgWindow_pnts]
		preMinVal = np.average(avgRange)

		# search backward from spike to find when vm reaches preMinVal (avg)
		# this could be a very long time if there is little to know hyper/re polarization
		# between spikes (like in an excitatory neuron)
		preRange = y[preMinPnt:spikeTimes[idx]]
		preRange = np.flip(preRange) # we want to search backwards from peak
		try:
			preMinPnt2 = np.where(preRange<preMinVal)[0][0]
			preMinPnt = spikeTimes[idx] - preMinPnt2
			
			preMinPntList[idx] = preMinPnt
			#spikeDict[iIdx]['preMinVal'] = preMinVal

		except (IndexError) as e:
			errorType = 'Pre spike min (mdp)'
			errorStr = 'Did not find preMinVal: ' + str(round(preMinVal,3)) #+ ' postRange min:' + str(np.min(postRange)) + ' max ' + str(np.max(postRange))
			eDict = getErrorDict(idx, spikeTimes[idx], errorType, errorStr) # spikeTime is in pnts
			errorList[idx].append(eDict)
			if verbose:
				logger.error(f'  spike:{idx} error:{eDict}')
			#preMinPntList.append(None)

	return preMinPntList, errorList

def getEarlyDiastolicDurationRate(y, spikeTimes : List[int],
							preMinPnts : List[int],
							lowEddRate_warning : float,
							verbose=False):
	"""Cardiac specific.

	Args:
		spikeTimes:
		preMinPnts (List[int]): Points where we found MDP.
			See getMaximumDepolaizingPotential()
	"""
	#
	# The nonlinear late diastolic depolarization phase was
	# estimated as the duration between 1% and 10% dV/dt
	# linear fit on 10% - 50% of the time from preMinPnt to self.spikeTimes[i]
	startLinearFit = 0.1 # percent of time between pre spike min and AP peak
	stopLinearFit = 0.5 #
	defaultVal = float('nan')

	x = np.arange(start=0, stop=len(y), step=1)

	retSpikeDict = [{}] * len(spikeTimes)
	errorList = [[]] * len(spikeTimes)

	for idx, spikeTime in enumerate(spikeTimes):
		preMinPnt = preMinPnts[idx]  # can be nan
		if preMinPnt is None:
			# TODO: If all are None then our return dict will be empty
			continue
		timeInterval_pnts = spikeTime - preMinPnt
		# taking round() so we always get an integer # points
		preLinearFitPnt0 = preMinPnt + round(timeInterval_pnts * startLinearFit)
		preLinearFitPnt1 = preMinPnt + round(timeInterval_pnts * stopLinearFit)
		preLinearFitVal0 = y[preLinearFitPnt0]
		preLinearFitVal1 = y[preLinearFitPnt1]

		# linear fit before spike
		retSpikeDict[idx]['preLinearFitPnt0'] = preLinearFitPnt0
		retSpikeDict[idx]['preLinearFitPnt1'] = preLinearFitPnt1
		#retSpikeDict[idx]['earlyDiastolicDuration_ms'] = self.pnt2Ms_(preLinearFitPnt1 - preLinearFitPnt0)
		retSpikeDict[idx]['preLinearFitVal0'] = preLinearFitVal0
		retSpikeDict[idx]['preLinearFitVal1'] = preLinearFitVal1

		# a linear fit where 'm,b = np.polyfit(x, y, 1)', m*x+b"
		xFit = x[preLinearFitPnt0:preLinearFitPnt1]
		yFit = y[preLinearFitPnt0:preLinearFitPnt1]

		# sometimes xFit/yFit have 0 length -->> TypeError
		#print(f' {iIdx} preLinearFitPnt0:{preLinearFitPnt0}, preLinearFitPnt1:{preLinearFitPnt1}')
		#print(f'    xFit:{len(xFit)} yFit:{len(yFit)}')

		with warnings.catch_warnings():
			warnings.filterwarnings('error')
			try:
				mLinear, bLinear = np.polyfit(xFit, yFit, 1) # m is slope, b is intercept
				retSpikeDict[idx]['earlyDiastolicDurationRate'] = mLinear
				# make an error if edd rate is too low
				if mLinear <= lowEddRate_warning:
					errorType = 'Fit EDD'
					errorStr = f'Early diastolic duration rate fit - Too low {round(mLinear,3)}<={lowEddRate_warning}'
					eDict = getErrorDict(idx, spikeTime, errorType, errorStr) # spikeTime is in pnts
					errorList[idx].append(eDict)
					if verbose:
						print(f'  spike:{idx} error:{eDict}')

			except (TypeError) as e:
				#catching exception:  expected non-empty vector for x
				# xFit/yFit turn up empty when mdp and TOP points are within 1 point
				retSpikeDict[idx]['earlyDiastolicDurationRate'] = defaultVal
				errorType = 'Fit EDD'
				errorStr = 'Early diastolic duration rate fit - preMinPnt == spikePnt'
				eDict = getErrorDict(idx, spikeTime, errorType, errorStr)
				errorList[idx].append(eDict)
				if verbose:
					logger.error(f'  spike:{idx} error:{eDict}')
			except (np.RankWarning) as e:
				#logger.error('== FIX preLinearFitPnt0/preLinearFitPnt1 RankWarning')
				#logger.error(f'  error is: {e}')
				#print('RankWarning')
				# also throws: RankWarning: Polyfit may be poorly conditioned
				retSpikeDict[idx]['earlyDiastolicDurationRate'] = defaultVal
				errorType = 'Fit EDD'
				errorStr = 'Early diastolic duration rate fit - RankWarning'
				eDict = getErrorDict(idx, spikeTime, errorType, errorStr)
				errorList[idx].append(eDict)
				if verbose:
					logger.error(f'  spike:{idx} error:{eDict}')
			except:
				logger.error(f' !!!!!!!! UNKNOWN EXCEPTION DURING EDD LINEAR FIT for spike {idx}')
				retSpikeDict[idx]['earlyDiastolicDurationRate'] = defaultVal
				errorType = 'Fit EDD'
				errorStr = 'Early diastolic duration rate fit - Unknown Exception'
				eDict = getErrorDict(idx, spikeTime, errorType, errorStr)
				errorList[idx].append(eDict)
				if verbose:
					logger.error(f'  spike:{idx} error:{eDict}')

	return retSpikeDict, errorList

def getErrorDict(spikeNumber, pnt, type, detailStr):
	"""
	Get error dictionary for one spike.
	"""
	#sec = self.pnt2Sec_(pnt)  # pnt / self.dataPointsPerMs / 1000
	#sec = round(sec,4)

	eDict = {
		'Spike': spikeNumber,
		'pnt': pnt,
		#'Seconds': sec,
		'Type': type,
		'Details': detailStr,
	}
	return eDict
	
def peakDetect(sweepY, detectionParams : sanpy.bDetection, dataPointsPerMs : int):
	"""Detect peaks in sweepY using specified detectionParams.
	
	Args:
		sweepY (np.ndarray): The signal to do detection on, usually a current-clamp Vm recording.
		detectionParams (sanpy.bDetection): Class that specifies the time-constants
		dataPointsPerMs (int):
	"""
	
	firstDerivative = getFirstDerivative(sweepY)

	verbose = detectionParams['verbose']

	#
	# pre-filter
	medianFilter = detectionParams['medianFilter']   # 0 means no filter
	if medianFilter > 0:
		sweepY = getMedian_Filter(sweepY, medianFilter, verbose)
	
	SavitzkyGolay_pnts = detectionParams['SavitzkyGolay_pnts']  # 0 means no filter
	SavitzkyGolay_poly = detectionParams['SavitzkyGolay_poly']
	if SavitzkyGolay_pnts > 0:
		sweepY = getSavitzkyGolay_Filter(sweepY, SavitzkyGolay_pnts, SavitzkyGolay_poly, verbose)
	
	#
	# detect peaks using different algorithms, either (dvdt, mv)
	detectionType = detectionParams['detectionType']  # in (dvdt, vm)

	if detectionType == 'dvdt':
		dvdtThreshold = detectionParams['dvdtThreshold']
		dvdtThreshold /= 10  # until I scale my derivative to sampling interval (10 pnts per ms is 10x slower)
		spikePnts = getThreshold_firstDerivative(sweepY, dvdtThreshold, firstDerivative=firstDerivative)
	elif detectionType == 'vm':
		mvThreshold = detectionParams['mvThreshold']
		spikePnts = getThreshold_vm(sweepY, mvThreshold)

	print(f'1 num spikes:{len(spikePnts)} are {spikePnts}')

	#
	# remove fast spikes
	refractory_pnts = detectionParams.getMsValueAsPnt('refractory_ms', dataPointsPerMs)
	spikePnts = reduceByRefractory(spikePnts, refractoryPnts=refractory_pnts)

	# refine spike times by looking in 1st and 2nd derivative
	# in general this is not working for (cardiac myocytes, DRG neurons) ???
	#dvdtPreWindow_pnts = detectionParams.getMsValueAsPnt('dvdtPreWindow_ms', dataPointsPerMs)
	#spikePnts = refineWithDerivative(sweepY, spikePnts, dvdtPreWindow_pnts,
	# 									firstDeriv=firstDerivative, secondDeriv=secondDerivative,
	# 									verbose=verbose)

	#
	# shift spike times to a percent of dvdt
	dvdt_percentOfMax = detectionParams['dvdt_percentOfMax']
	spikePnts, errorList = refineWithPercentOfMax(sweepY, spikePnts, dvdt_percentOfMax,
								firstDeriv=firstDerivative, verbose=verbose)

	#
	# remove based on (onlyPeaksAbove_mV, onlyPeaksBelow_mV)
	onlyPeaksAbove_mV = detectionParams['onlyPeaksAbove_mV']
	spikePnts = keepPeaksAbove(sweepY, spikePnts, onlyPeaksAbove_mV, verbose=verbose)

	onlyPeaksBelow_mV = detectionParams['onlyPeaksBelow_mV']
	spikePnts = keepPeaksBelow(sweepY, spikePnts, onlyPeaksBelow_mV, verbose=verbose)

	###
	# now that we have a good threshold crossing, calculate everything else
	# number in peakPnts will not change
	###
	print(f'2 num spikes:{len(spikePnts)} are {spikePnts}')

	# peak
	peakWindow_pnts = detectionParams.getMsValueAsPnt('peakWindow_ms', dataPointsPerMs)
	spikePeakPnts = getPeaks(sweepY, spikePnts, peakWindow_pnts)

	# half-width
	halfWidthWindow_pnts = detectionParams.getMsValueAsPnt('halfWidthWindow_ms', dataPointsPerMs)
	halfHeights = detectionParams['halfHeights']
	
	hwList, hwErrorList = getHalfWidth(sweepY, spikePnts, spikePeakPnts, 
							halfWidthWindow_pnts, dataPointsPerMs,
							halfHeights=halfHeights, verbose=verbose)
	
	# maximum diastolic potential (MDP) between spikes
	mdp_pre_pnts = detectionParams.getMsValueAsPnt('mdp_ms', dataPointsPerMs)
	avgWindow_pnts = detectionParams.getMsValueAsPnt('avgWindow_ms', dataPointsPerMs)

	mdpList, mdpErrorList = getMaximumDepolaizingPotential(sweepY, spikePnts,
				mdp_pre_pnts=mdp_pre_pnts, avgWindow_pnts=avgWindow_pnts,
				verbose=verbose)

	lowEddRate_warning = detectionParams['lowEddRate_warning']
	eddRateList, eddRateErrorList = getEarlyDiastolicDurationRate(sweepY, spikePnts, mdpList,
									lowEddRate_warning=lowEddRate_warning,
									verbose=verbose)

	resultsDict = {
		'spikeTimes' : spikePnts,
		'spikePeaks' : spikePeakPnts,
		'widths' : hwList,
	}

	return resultsDict

def peakDetect_ba(ba, sweep, detectionParams):
	"""Detect one sweep in a bAnalysis.
	"""
	ba.setSweep(sweep)
	
	sweepY = ba.sweepY
	dataPointsPerMs = ba.dataPointsPerMs

	resultsDict = peakDetect(sweepY, detectionParams, dataPointsPerMs)

	return resultsDict

def testNeuron():
	path = '/Users/cudmore/data/theanne-griffith/07.20.21/2021_07_20_0013.abf'
	ba = sanpy.bAnalysis(path)
	print(ba)

	detectionPreset = sanpy.bDetection.detectionPresets.neuron
	detectionParams = sanpy.bDetection(detectionPreset=detectionPreset)

	for sweep in range(ba.numSweeps):
		resultsDict = peakDetect_ba(ba, sweep, detectionParams)

		sweepX = ba.sweepX
		sweepX_pnts = np.arange(start=0, stop=sweepX.shape[0], step = 1)
		sweepY = ba.sweepY

		Plot(sweepX_pnts, sweepY, resultsDict)
		plt.show()

def _plotHalfWidth(y, hwSpikes, ax):
	"""
	
	Args:
		hwSpikes (list): a list of spikes, each has a list of hw dict.
	"""
	
	for oneSpikeHwList in hwSpikes:
		for oneHw in oneSpikeHwList:
			# oneHw is a dict
			# {'halfHeight': 10, 'risingPnt': 1579, 'fallingPnt': 1590, 'widthPnts': 11, 'widthMs': 1.1}
			#print('  oneHw:', oneHw)
			
			risingPnt = oneHw['risingPnt']
			fallingPnt = oneHw['fallingPnt']

			if risingPnt is None or fallingPnt is None:
				# error, we did not find a rising/falling point
				continue

			yRisingPnt = y[risingPnt]
			yFallingPnt = y[fallingPnt]

			ax.plot(risingPnt, yRisingPnt, 'o')
			ax.plot(fallingPnt, yFallingPnt, 'o')
		

def Plot(x, y, resultsDict : dict ={}, showFirstDeriv=True, showSecondDeriv=True):
	
	if showFirstDeriv:
		firstDerivative = getFirstDerivative(y)
	if showSecondDeriv:
		secondDerivative = getSecondDerivative(y)

	if showFirstDeriv or showSecondDeriv:
		numRows = 2
		yAxisIdx = 1
	else:
		numRows = 1
		yAxisIdx = 0

	fig, axs = plt.subplots(numRows, 1, sharex=True)

	if showFirstDeriv:
		axs[0].plot(x, firstDerivative, 'k', linewidth=1)
		axs[0].hlines(0, x[0], x[-1], linestyles='dashed')
	if showSecondDeriv:
		rightAxes=axs[0].twinx()
		rightAxes.plot(x, secondDerivative, 'r', linewidth=1)
		rightAxes.hlines(0, x[0], x[-1], linestyles='dashed')

	axs[yAxisIdx].plot(x, y, 'k', linewidth=1)

	if resultsDict is not None:
		#
		spikeTimes = resultsDict['spikeTimes']
		spikeVals = y[spikeTimes]
		axs[yAxisIdx].plot(spikeTimes, spikeVals, 'or')
		#
		spikePeaks = resultsDict['spikePeaks']
		spikePeaksVals = y[spikePeaks]
		axs[yAxisIdx].plot(spikePeaks, spikePeaksVals, 'ob')

		_plotHalfWidth(y, resultsDict['widths'], axs[yAxisIdx])
		#for idx, spikeTime in enumerate(spikeTimes):
		#	# ['widhts'] is a list of dict
		#	print(f"idx:{idx} width: {resultsDict['widths'][idx]}")

	#plt.show()

if __name__ == '__main__':
	import matplotlib.pyplot as plt
	import sanpy
	
	testNeuron()


from typing import List

from PyQt5 import QtCore, QtGui, QtWidgets

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class errorTableView(QtWidgets.QTableView):
    """
    Display a per spike error table (one row per spike eror)
    """
    signalSelectSpike = QtCore.Signal(object) # spike number, doZoom

    def __init__(self, parent=None):
        super(errorTableView, self).__init__(parent)

        #self.setFont(QtGui.QFont('Arial', 10))

        self._blockTableUpdate = False

        self.setSortingEnabled(True)

        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                  QtWidgets.QSizePolicy.Expanding)
        
        self.setSelectionBehavior(QtWidgets.QTableView.SelectRows)
        
        #self.setSelectionMode(QtWidgets.QTableView.SingleSelection)

        self.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers |
                             QtWidgets.QAbstractItemView.DoubleClicked)

        # equally stretchs each columns so that they fit the table's width.
        #self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch);
        self.horizontalHeader().setStretchLastSection(True)

        # removed mar 2023
        # rowHeight = 11
        # fnt = self.font()
        # fnt.setPointSize(rowHeight)
        # self.setFont(fnt)
        # self.verticalHeader().setDefaultSectionSize(rowHeight);

        #self.resizeColumnsToContents()

        self.clicked.connect(self.errorTableClicked)

    def mySelectRows(self, rows : List[int]):
        """Select rows in the table
        
        TODO: Hold off on this, we need to define a sort proxy. See PyMapManager tables.
        """
        return
    
        logger.info(f'rows:{rows}')

        # to stop event recursion
        self._blockTableUpdate = True
        
        selectionModel = self.selectionModel()
        if selectionModel:
            selectionModel.clear()
        
            if rows:
                indexes = [self.model().index(r, 0) for r in rows]  # [QModelIndex]
                visualRows = [self.proxy.mapFromSource(modelIndex) for modelIndex in indexes]

                mode = QtCore.QItemSelectionModel.Select | QtCore.QItemSelectionModel.Rows
                [self.selectionModel().select(i, mode) for i in visualRows]

                # scroll so first row in rows is visible
                # TODO (cudmore) does not work if list is filtered
                column = 0
                row = list(rows)[0]
                index = self.model().index(row, column)
                self.scrollTo(index, QtWidgets.QAbstractItemView.PositionAtTop)  # EnsureVisible
            else:
                #print('  CLEARING SELECTION')
                self.clearSelection()
        
        #
        self._blockTableUpdate = False
        
    def errorTableClicked(self, index):
        row = self.model()._data.index[index.row()] # take sorting into account

        doZoom = False
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ShiftModifier:
            # zoomm on shift+click
            doZoom = True

        try:
            # .loc[i,'col'] gets from index label (correct)
            # .iloc[i,j] gets absolute row (wrong)
            spikeNumber = self.model()._data.loc[row, 'Spike']
        except (KeyError) as e:
            # for results plugin
            try:
                spikeNumber = self.model()._data.loc[row, 'spikeNumber']
            except (KeyError) as e:
                logger.warning(f'KeyError looking for column "Spike" or "spikeNumber"')
                return
        spikeNumber = int(spikeNumber)

        dDict = {
            'spikeNumber': spikeNumber,
            'doZoom': doZoom,
        }
        logger.info(f'  -->> emit signalSelectSpike with dDict')
        logger.info(dDict)
        
        #self._blockSlots = True
        self.signalSelectSpike.emit(dDict)
        
        #self._blockSlots = False




import os
import pathlib
import json
from json.decoder import JSONDecodeError

import sanpy
import sanpy.interface
import sanpy.interface.sanpy_app as sanpy_app

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class preferences():
    """Class to hold GUI preferences.
    
    Created and used by sanpy app SanPyWindow.
    """
    def __init__(self, sanpyApp : sanpy_app):
        super().__init__()
        self._sanpyApp = sanpyApp
        
        # bump this when we change, sanpy app will rebuild if loaded preerences are out of date
        #self._version = 1.2  # increment when we change preferences
        self._version = 1.3  # added set spike
        self._version = 1.4  # converted plugins from list[str] to dict to include externalWindow option
        self._version = 1.5  # adding keys to plugins (l, t, w, h)
        
        self._maxRecent = 7  # a lucky number
        self._configDict = self.load()

    def __setitem__(self, key, item):
        self._configDict[key] = item

    def __getitem__(self, key):
        return self._configDict[key]

    @property
    def configDict(self):
        return self._configDict
    
    def addFolder(self, path : str):
        """Add a folder to preferences.
        """
        logger.info(f'{path}')
        
        # add if not in recentFolders
        if not path in self.configDict['recentFolders']:
            self.configDict['recentFolders'].append(path)
            # limit list to last _maxNumUndo 
            self.configDict['recentFolders'] = self.configDict['recentFolders'][-self._maxRecent:] 
    
        # always set as the most recent folder
        self.configDict['mostRecentFolder'] = path

    def addPlugin(self, pluginName : str, externalWindow : bool,
                  ltwhTuple : tuple):
        """Add plugin to preferences.
        """
        # was this as list
        # if not pluginName in self.configDict['pluginPanels']:
        #     self.configDict['pluginPanels'].append(pluginName)
        if not pluginName in self.configDict['pluginPanels'].keys():
            self.configDict['pluginPanels'][pluginName] = {'externalWindow': externalWindow,
                                                           'l': ltwhTuple[0],
                                                           't': ltwhTuple[1],
                                                           'w': ltwhTuple[2],
                                                           'h': ltwhTuple[3],
                                                           }

    def removePlugin(self, pluginName : str):
        """Remove plugin from preferences.
        """
        # was this as list
        # if pluginName in self.configDict['pluginPanels']:
        #     self.configDict['pluginPanels'].remove(pluginName)
        if pluginName in self.configDict['pluginPanels'].keys():
            self.configDict['pluginPanels'].pop(pluginName)

    def getMostRecentFolder(self):
        return self.configDict['mostRecentFolder']
        
    def getRecentFolder(self):
        return self.configDict['recentFolders']

    def preferencesSet(self, key1, key2, val):
        """Set a preference. See `getDefaults()` for key values."""
        try:
            self._configDict[key1][key2] = val

            # actually show hide some widgets
            #if key1=='display' and key2=='showScatter':
            #    if val:
            #        self.myScatterPlotWidget.show()
            #    else:
            #        self.myScatterPlotWidget.hide()

            #if key1=='display' and key2=='showErrors':
            #    if val:
            #        self.myErrorTable.show()
            #    else:
            #        self.myErrorTable.hide()

        except (KeyError) as e:
            logger.error(f'Did not set preference with keys "{key1}" and "{key2}"')

    def preferencesGet(self, key1, key2):
        """Get a preference. See `getDefaults()` for key values."""
        try:
            return self._configDict[key1][key2]
        except (KeyError) as e:
            logger.error(f'Did not get preference with keys "{key1}" and "{key2}"')

    def getPreferencesFile(self):
        userPreferencesFolder = sanpy._util._getUserPreferencesFolder()
        optionsFile = pathlib.Path(userPreferencesFolder) / 'sanpy_preferences.json'
        return optionsFile

    def load(self):
        """
        Always load preferences from:
            <user>/Documents/SanPy/preferences/sanpy_preferences.json
        """

        preferencesFile = self.getPreferencesFile()

        useDefault = True
        if os.path.isfile(preferencesFile):
            logger.info(f'Loading preferences file')
            logger.info(f'  {preferencesFile}')
            try:
                with open(preferencesFile) as f:
                    loadedJson = json.load(f)
                    loadedVersion = loadedJson['version']
                    logger.info(f'  loaded preferences version {loadedVersion}')
                    if loadedVersion < self._version:
                        # use default
                        logger.warning('  older version found, reverting to current defaults')
                        logger.warning(f'  loadedVersion:{loadedVersion} currentVersion:{self._version}')
                        pass
                    else:
                        return loadedJson
            except (JSONDecodeError) as e:
                logger.error(e)
            except (TypeError) as e:
                logger.error(e)
        if useDefault:
            logger.info(f'Using default options')
            return self.getDefaults()

    def getDefaults(self) -> dict:
        """Get default preferences.

        Be sure to increment self._version when making changes.
        """
        configDict = {}  # OrderedDict()

        configDict['version'] = self._version
        configDict['useDarkStyle'] = True
        configDict['autoDetect'] = True # FALSE DOES NOT WORK!!!! auto detect on file selection and/or sweep selection

        configDict['recentFolders'] = []
        configDict['mostRecentFolder'] = ''

        configDict['windowGeometry'] = {}
        configDict['windowGeometry']['x'] = 100
        configDict['windowGeometry']['y'] = 100
        configDict['windowGeometry']['width'] = 1000
        configDict['windowGeometry']['height'] = 1000
        
        configDict['filePanels'] = {}
        configDict['filePanels']['File Panel'] = True

        # panels withing detectionWidget -> myDetectionToolbar
        configDict['detectionPanels'] = {}
        configDict['detectionPanels']['Detection Panel'] = True  # main panel
        configDict['detectionPanels']['Detection'] = True
        configDict['detectionPanels']['Display'] = True
        configDict['detectionPanels']['Plot Options'] = False
        configDict['detectionPanels']['Set Spikes'] = False
        
        # plugins to show at startup
        # 20230325 convert from List[str] to dict
        configDict['pluginPanels'] = {}

        # values in detectionWidget -> myDetectionToolbar
        configDict['detect'] = {}
        configDict['detect']['detectDvDt'] = 20
        configDict['detect']['detectMv'] = -20

        # display options within detectionWidget -> myDetectionToolbar
        configDict['rawDataPanels'] = {}
        configDict['rawDataPanels']['plotEveryPoint'] = 10 # not used?
        configDict['rawDataPanels']['Full Recording'] = True #
        configDict['rawDataPanels']['Derivative'] = False #
        configDict['rawDataPanels']['DAC'] = False #

        return configDict

    def save(self):
        preferencesFile = self.getPreferencesFile()
        
        logger.info(f'Saving preferences file as: "{preferencesFile}"')

        # get the current position and size of the main window
        left, top, width, height = self._sanpyApp.getWindowGeometry()
        
        self._configDict['windowGeometry']['x'] = left
        self._configDict['windowGeometry']['y'] = top
        self._configDict['windowGeometry']['width'] = width
        self._configDict['windowGeometry']['height'] = height

        # TODO: extend to a list
        #self._configDict['lastPath'] = self._sanpyApp.path

        # TODO: get window position and size of all open plugins
        # self.configDict['pluginPanels'][pluginName] = {'externalWindow': externalWindow,
        #                                                    'l': ltwhTuple[0],
        #                                                    't': ltwhTuple[1],
        #                                                    'w': ltwhTuple[2],
        #                                                    'h': ltwhTuple[3],
        #                                                    }
        _openSet = self._sanpyApp.myPlugins._openSet
        for _onePlugin in _openSet:
            pluginName = _onePlugin.myHumanName
            ltwhTuple = _onePlugin.getWindowGeometry()
            self.configDict['pluginPanels'][pluginName]['l'] = ltwhTuple[0]
            self.configDict['pluginPanels'][pluginName]['t'] = ltwhTuple[1]
            self.configDict['pluginPanels'][pluginName]['w'] = ltwhTuple[2]
            self.configDict['pluginPanels'][pluginName]['h'] = ltwhTuple[3]

        #
        # save
        with open(preferencesFile, 'w') as outfile:
            json.dump(self._configDict, outfile, indent=4, sort_keys=True)

        self._sanpyApp.slot_updateStatus('Saved preferences')


#import os, sys;
#sys.path.append(os.path.dirname(os.path.realpath(__file__)))

#from . import *

from .sanpy_app import SanPyWindow

from .bDetectionWidget import bDetectionWidget
from .bScatterPlotWidget2 import bScatterPlotMainWindow
from .bScatterPlotWidget2 import myTableView as myTableView_tmp  # used by fi plugin
from .bExportWidget import bExportWidget
from .bFileTable import *

from .bTableView import bTableView
from .fileListWidget import fileListWidget

from .bErrorTable import *
from .bDialog import *
#from .bFileTable import *
from .bPlugins import *
#from .aPlugin import *

# critical for plugins to be loaded dynamically
#from .plugins import *

#from .breeze_resources import *

from .bKymographWidget import kymographWidget
from .kymographPlugin2 import kymographPlugin2

from .preferences import preferences


import pandas as pd

from PyQt5 import QtWidgets

import sanpy
from sanpy.interface.plugins import sanpyPlugin

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

class detectionErrors(sanpyPlugin):
    """
    Plugin to display detection errors

    Uses:
        QTableView: sanpy.interface.bErrorTable.errorTableView()
        QAbstractTableModel: sanpy.interface.bFileTable.pandasModel
    """
    myHumanName = 'Summary Error'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        #self.pyqtWindow() # makes self.mainWidget

        self._dfError = None

        layout = QtWidgets.QVBoxLayout()

        self.myErrorTable = sanpy.interface.bErrorTable.errorTableView()
        self.myErrorTable.setWordWrap(False)

        layout.addWidget(self.myErrorTable)

        #self.setLayout(layout)
        self.getVBoxLayout().addLayout(layout)
        
        #
        # connect clicks in error table to siganl main sanpy_app with slot_selectSpike()
        logger.warning('mar 11 FIX SPIKE SELECTION')
        if self.getSanPyApp() is not None:
            fnPtr = self.getSanPyApp().slot_selectSpike
            self.myErrorTable.signalSelectSpike.connect(fnPtr)

        self.replot()

    def replot(self):
        # update
        # todo: get default columns from ???
        self._dfError = pd.DataFrame(columns=['Spike', 'Seconds', 'Type', 'Details'])

        if self.ba is not None:
            if self.ba.dfError is not None:
                self._dfError = self.ba.dfError
        #
        if self._dfError is not None:
            errorReportModel = sanpy.interface.bFileTable.pandasModel(self._dfError)
            self.myErrorTable.setModel(errorReportModel)
        else:
            logger.error('Did not get error df from bAnalysis')

    def copyToClipboard(self):
        if self._dfError is not None:
            self._dfError.to_clipboard(sep='\t', index=False)
            logger.info('Error table copied to clipboard')

if __name__ == '__main__':
    import sys
    app = QtWidgets.QApplication([])
    spl = detectionErrors()
    spl.show()
    sys.exit(app.exec_())



import os
import sys
import time
import pathlib
import shutil

import pandas as pd

import tables.scripts.ptrepack  # to save compressed .h5 file

#import logging
#logging.getLogger('numexpr').setLevel(logging.WARNING)
from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

def listKeys(hdfPath, printData=False):
    """List all keys in h5 file."""
    with pd.HDFStore(hdfPath, mode='r') as store:
        print(store)
        for key in store.keys():
            print('    key:', key, 'type:', type(store[key]))
            if printData:
                print(store[key])

def _getTmpHdfFile(hdfPath):
    """Get temporary h5 file to write to.
    
    We will always then compress with _rebuildHdf.

    Used in analysisDir to delete uuid from h5 file
    """
    if not os.path.isfile(hdfPath):
        logger.error(f'File not found {hdfPath}')
        return

    logger.info('')
    
    _folder, _file = os.path.split(hdfPath)
    
    tmpHdfFile = os.path.splitext(_file)[0] + '_tmp.h5'
    tmpHdfPath = pathlib.Path(_folder) / tmpHdfFile

    # the compressed version from the last save
    hdfFile = os.path.splitext(_file)[0] + '.h5'
    hdfFilePath = pathlib.Path(_folder) / hdfFile
    
    #hdfMode = 'w'
    if os.path.isfile(hdfFilePath):
        logger.info(f'    copying existing hdf file to tmp ')
        logger.info(f'    hdfFilePath {hdfFilePath}')
        logger.info(f'    tmpHdfPath {tmpHdfPath}')
        shutil.copyfile(hdfFilePath,tmpHdfPath)
    else:
        pass
        # compressed file does not exist, just use tmp path
        # print('   does not exist:', hdfFilePath)

    return tmpHdfPath

def _repackHdf(hdfPath):

    _folder, _file = os.path.split(hdfPath)

    # rebuild the file to remove old changes and reduce size
    #tmpHdfFile = os.path.splitext(_file)[0] + '_tmp.h5'
    #tmpHdfPath = pathlib.Path(_folder) / tmpHdfFile
    
    # always write to a tmp and then repack into hdfPath
    tmpHdfPath = _getTmpHdfFile(hdfPath)

    hdfFile = os.path.splitext(_file)[0] + '.h5'
    hdfPath = pathlib.Path(_folder) / hdfFile
    logger.info(f'Rebuilding h5 to {hdfPath}')
    
    # can't pass sys.argv a 'PosixPath' from pathlib.Path, needs to be a string
    tmpHdfPath = str(tmpHdfPath)
    hdfPath = str(hdfPath)
    
    # when calling ptrepack, we need trailing ':' on each src/dst path
    # without this Windows fails to find the file
    _tmpHdfPath = tmpHdfPath + ':'
    _hdfPath = hdfPath + ':'
    
    # The first item is normally the command line command name (not used)
    sys.argv = ["", "--overwrite", "--chunkshape=auto", _tmpHdfPath, _hdfPath]

    logger.info('    running tables.scripts.ptrepack.main()')
    logger.info(f'    sys.argv: {sys.argv}')
    try:
        
        tables.scripts.ptrepack.main()

        # delete the temporary (large file)
        logger.info(f'    Deleting tmp file: {tmpHdfPath}')
        os.remove(tmpHdfPath)

        #self.signalApp(f'Saved compressed folder analysis with tables.scripts.ptrepack.main()')

    except(FileNotFoundError) as e:
        logger.error('tables.scripts.ptrepack.main() failed ... file was not saved')
        logger.error(e)
        #self.signalApp(f'ERROR in tables.scripts.ptrepack.main(): {e}')

def _loadAnalysis(hdfPath):
    """Load all bAnalysis from h5"""
    logger.info(f'hdfPath: {hdfPath}')
    start = time.time()
    numLoaded =0
    with pd.HDFStore(hdfPath, mode='r') as store:
        for key in store.keys():
            data = store[key]
            if isinstance(data, pd.DataFrame):
                if '_sweepX' in data.columns:
                    ba = sanpy.bAnalysis(fromDf=data)
                    logger.info(f'loaded ba: {ba}')
                    numLoaded += 1
                else:
                    # this is usually the file database
                    print('Dur(s):')
                    print(store[key]['Dur(s)'])
                    # the entire file db df
                    print(data)
    #
    stop = time.time()
    print(f'Loading {numLoaded} bAnalysis took {round(stop-start,3)} seconds.')

if __name__ == '__main__':
    import sanpy.bAnalysis

    hdfPath = '/home/cudmore/Sites/SanPy/data/sanpy_recording_db.h5'
    #listKeys(hdfPath)
    _loadAnalysis(hdfPath)

    # the file befor compression
    '''
    hdfPath = '/home/cudmore/Sites/SanPy/data/sanpy_recording_db_tmp.h5'
    loadAnalysis(hdfPath)
    '''



import os
import sys
import importlib
import pathlib
import shutil
from typing import List, Union
import uuid

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

def getNewUuid():
    return 't' + str(uuid.uuid4()).replace('-', '_')

def getBundledDir():
    """Get the working directory where user preferences are save.

    This will be source code folder when running from source,
      will be a more freeform folder when running as a frozen app/exe
    """
    if getattr(sys, 'frozen', False):
        # we are running in a bundle (frozen)
        bundle_dir = sys._MEIPASS
    else:
        # we are running in a normal Python environment
        bundle_dir = os.path.dirname(os.path.abspath(__file__))    
    return bundle_dir

def _module_from_file(module_name : str, file_path : str):
    """
    
    Args:
        module_name: Is like sanpy.interface.plugins.onePluginFile
        file_path: Full path to onePluginFile source code (onePluginFile.py)
    """
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module

def addUserPath():
    """Make <user>/Documents/SanPy folder and add it to the Python sys.path
    
    Returns:
        True: If we made the folder (first time SanPy is running)
    """
    
    logger.info('')

    madeUserFolder = _makeSanPyFolders()  # make <user>/Documents/SanPy if necc
            
    userSanPyFolder = _getUserSanPyFolder()

    # if userSanPyFolder in sys.path:
    #     sys.path.remove(userSanPyFolder)

    if not userSanPyFolder in sys.path:
        logger.info(f'Adding to sys.path: {userSanPyFolder}')
        sys.path.append(userSanPyFolder)

    logger.info('sys.path is now:')
    for path in sys.path:
        logger.info(f'    {path}')

    return madeUserFolder
    
def _getUserDocumentsFolder():
    """Get <user>/Documents folder.
    """
    userPath = pathlib.Path.home()
    userDocumentsFolder = os.path.join(userPath, 'Documents')
    if not os.path.isdir(userDocumentsFolder):
        logger.error(f'Did not find path "{userDocumentsFolder}"')
        logger.error(f'   Using "{userPath}"')
        return userPath
    else:
        return userDocumentsFolder

def _getUserSanPyFolder():
    """Get <user>/Documents/SanPy folder.
    """
    userDocumentsFolder = _getUserDocumentsFolder()
    sanpyFolder = os.path.join(userDocumentsFolder, 'SanPy')
    return sanpyFolder

def old_getUserFolder(folder : str) -> str:
    userSanPyFolder = _getUserSanPyFolder()
    if folder == 'plugins':
        theFolder = os.path.join(userSanPyFolder, 'plugins')
    elif folder == 'analysis':
        theFolder = os.path.join(userSanPyFolder, 'analysis')
    elif folder == 'preferences':
        theFolder = os.path.join(userSanPyFolder, 'preferences')
    else:
        logger.error(f'did not understand folder: "{folder}"')
        return None
    return theFolder

def _getUserFileLoaderFolder():
    userSanPyFolder = _getUserSanPyFolder()
    fileLoaderFolder = os.path.join(userSanPyFolder, 'file loaders')
    return fileLoaderFolder

def _getUserPluginFolder():
    userSanPyFolder = _getUserSanPyFolder()
    userPluginFolder = os.path.join(userSanPyFolder, 'plugins')
    return userPluginFolder

def _getUserDetectionFolder():
    """Folder of saved user detection presets.
    
    Each is a json file.
    """
    userSanPyFolder = _getUserSanPyFolder()
    userDetectionFolder = os.path.join(userSanPyFolder, 'detection')
    return userDetectionFolder

def _getUserAnalysisFolder():
    """Folder of custom user analysis code.
    """
    userAnalysisFolder = _getUserSanPyFolder()
    userAnalysisFolder = os.path.join(userAnalysisFolder, 'analysis')
    return userAnalysisFolder

def _getUserPreferencesFolder():
    """Folder of SanPy app preferences and logs (user does not modify this.
    """
    userPreferencesFolder = _getUserSanPyFolder()
    userPreferencesFolder = os.path.join(userPreferencesFolder, 'preferences')
    return userPreferencesFolder

def pprint(d : dict):
    for k,v in d.items():
        print(f'  {k}: {v}')

def _makeSanPyFolders():
    """Make <user>/Documents/SanPy folder .

    If no Documents folder then make SanPy folder directly in <user> path.
    """
    userDocumentsFolder = _getUserDocumentsFolder()

    madeUserFolder = False
    
    # main <user>/DOcuments/SanPy folder
    sanpyFolder = _getUserSanPyFolder()
    if not os.path.isdir(sanpyFolder):
        # first time run
        logger.info(f'Making <user>/SanPy folder "{sanpyFolder}"')
        madeUserFolder = True
        #
        # copy entire xxx into <user>/Documents/SanPy
        _bundDir = getBundledDir()
        _srcPath = pathlib.Path(_bundDir) / '_userFiles' / 'SanPy'
        _dstPath = pathlib.Path(sanpyFolder)
        logger.info(f'    copying folder tree to <user>/Documents/SanPy folder')
        logger.info(f'    _srcPath:{_srcPath}')
        logger.info(f'    _dstPath:{_dstPath}')
        shutil.copytree(_srcPath, _dstPath)
    else:
        # already exists, make sure we have all sub-folders that are expected
        pass

    return madeUserFolder

    '''
    # SanPyapplication preferences
    preferencesFolder = _getUserPreferencesFolder()
    if not os.path.isdir(preferencesFolder):
        logger.info(f'Making user preferences folder "{preferencesFolder}"')
        os.mkdir(preferencesFolder)
    
    # to save detection parameters json
    detectionFolder = os.path.join(sanpyFolder, 'detection')
    if not os.path.isdir(detectionFolder):
        logger.info(f'Making user detection folder "{detectionFolder}"')
        os.mkdir(detectionFolder)

    # to hold custom .py plugins
    pluginFolder = _getUserPluginFolder()  # os.path.join(sanpyFolder, 'plugins')
    if not os.path.isdir(pluginFolder):
        # TODO: add __init__.py
        logger.info(f'Making user plugin folder "{pluginFolder}"')
        os.mkdir(pluginFolder)

    # to hold custom .py analysis
    analysisFolder = _getUserAnalysisFolder()
    if not os.path.isdir(detectionFolder):
        # TODO: add __init__.py
        logger.info(f'Making user analysis folder "{analysisFolder}"')
        os.mkdir(analysisFolder)
        # copy from _userFiles into this folder
        # _bundDir = getBundledDir()
        # _srcPath = pathlib.Path(self._bundDir) / tmpHdfFile
        # shutil.copyfile(hdfFilePath,tmpHdfPath)
    '''

# def _module_from_file(module_name, file_path):
# 	spec = importlib.util.spec_from_file_location(module_name, file_path)
# 	module = importlib.util.module_from_spec(spec)
# 	spec.loader.exec_module(module)
# 	return module

def _loadLineScanHeader(path):
    """
    path: full path to tif

    we will load and parse coresponding .txt file

    returns dict:
        numPixels:
        umLength:
        umPerPixel:
        totalSeconds:
    """
    # "X Dimension"	"138, 0.0 - 57.176 [um], 0.414 [um/pixel]"
    # "T Dimension"	"1, 0.000 - 35.496 [s], Interval FreeRun"
    # "Image Size(Unit Converted)"	"57.176 [um] * 35500.000 [ms]"

    # 20220606, adding
    # "Image Size"	"294 * 1000 [pixel]"

    txtFile = os.path.splitext(path)[0] + '.txt'

    if not os.path.isfile(txtFile):
        logger.error(f'did not find file:{txtFile}')
        return None

    theRet = {'tif': path}

    # tif shape is (lines, pixels)
    #theRet['numLines'] = self.tif.shape[1]
    #theRet['numLines'] = tifData.shape[0]

    gotImageSize = False

    with open(txtFile, 'r') as fp:
        lines = fp.readlines()
        for line in lines:
            line = line.strip()
            if line.startswith('"X Dimension"'):
                line = line.replace('"', "")
                line = line.replace(',', "")
                #print('loadLineScanHeader:', line)
                # 2 number of pixels in line
                # 5 um length of line
                # 7 um/pixel
                splitLine = line.split()
                for idx, split in enumerate(splitLine):
                    #print('  ', idx, split)
                    if idx == 2:
                        numPixels = int(split)
                        theRet['numPixels'] = numPixels
                    elif idx == 5:
                        umLength = float(split)
                        theRet['umLength'] = umLength
                    elif idx == 7:
                        umPerPixel = float(split)
                        theRet['umPerPixel'] = umPerPixel

            elif line.startswith('"T Dimension"'):
                # "T Dimension"	"1, 0.000 - 35.496 [s], Interval FreeRun"
                line = line.replace('"', "")
                line = line.replace(',', "")
                #print('loadLineScanHeader:', line)
                # 5 total duration of image acquisition (seconds)
                splitLine = line.split()
                for idx, split in enumerate(splitLine):
                    #print('  ', idx, split)
                    if idx == 5:
                        totalSeconds = float(split)
                        theRet['totalSeconds'] = totalSeconds

                        #theRet['secondsPerLine'] =

            # order in file will matter, there are multiple "Image Size" lines
            # we want the first
            # "Image Size"	"294 * 1000 [pixel]"
            elif line.startswith('"Image Size"'):
                if line.startswith('"Image Size(Unit Converted)"'):
                    continue
                if gotImageSize:
                    continue
                gotImageSize = True
                line = line.replace('"', "")
                line = line.replace(',', "")
                splitLine = line.split('\t')  # yes, a FREAKING tab !!!!
                splitLine = splitLine[1]
                splitLine2 = splitLine.split()
                #print('splitLine2:', splitLine2)

                theRet['numLines'] = int(splitLine2[2])

            #elif line.startswith('"Image Size(Unit Converted)"'):
            #	print('loadLineScanHeader:', line)

    # tif shape is (lines, pixels)
    shape = (theRet['numLines'], theRet['numPixels'])
    #theRet['shape'] = self.tif.shape
    #theRet['shape'] = tifData.shape
    theRet['shape'] = shape
    theRet['secondsPerLine'] = theRet['totalSeconds'] / theRet['shape'][0]
    theRet['linesPerSecond'] = 1 / theRet['secondsPerLine']
    #
    return theRet



import numpy as np
import scipy.signal

from PyQt5 import QtCore, QtWidgets, QtGui

import matplotlib.pyplot as plt

from sanpy.sanpyLogger import get_logger
logger = get_logger(__name__)

import sanpy
from sanpy.interface.plugins import sanpyPlugin

from sanpy.bAnalysisUtil import statList

class exportTrace(sanpyPlugin):
    """
    """
    myHumanName = 'Export Trace'

    def __init__(self, **kwargs):
        super(exportTrace, self).__init__(**kwargs)

        # this plugin will not respond to any interface changes
        #self.turnOffAllSignalSlot()

        self.mainWidget = None

        self.plot()

    def replot(self):
        logger.info(f'sweepNumber: {self.sweepNumber}')

        if self.mainWidget is None:
            self.plot()

        x = self.getSweep('x')  # get for current sweep
        y = self.getSweep('y')
        self.mainWidget.switchFile(self.ba, x, y)
        
    def plot(self):
        if self.ba is None:
            return

        self.myType = 'vmFiltered'

        if self.myType == 'vmFiltered':
            xyUnits = ('Time (sec)', 'Vm (mV)')# todo: pass xMin,xMax to constructor
            x = self.getSweep('x')  # corresponds to sweepX
            y = self.getSweep('y')  # corresponds to sweepY
        elif self.myType == 'dvdtFiltered':
            xyUnits = ('Time (sec)', 'dV/dt (mV/ms)')# todo: pass xMin,xMax to constructor
        elif self.myType == 'meanclip':
            xyUnits = ('Time (ms)', 'Vm (mV)')# todo: pass xMin,xMax to constructor
        else:
            logger.error(f'Unknown myType: "{self.myType}"')
            xyUnits = ('error time', 'error y')

        path = self.ba.fileLoader.filepath

        xMin, xMax = self.getStartStop()
        '''
        xMin = None
        xMax = None
        if self.myType in ['clip', 'meanclip']:
            xMin, xMax = self.detectionWidget.clipPlot.getAxis('bottom').range
        else:
            xMin, xMax = self.detectionWidget.getXRange()
        '''

        if self.myType in ['vm', 'dvdt']:
            xMargin = 2 # seconds
        else:
            xMargin = 2

        # a large x margin was causing matplotlib erors when a swep was short, like 0.2 sec
        xMargin = 0
        
        tmpLayout = QtWidgets.QVBoxLayout()

        self.mainWidget = sanpy.interface.bExportWidget(x, y,
                        xyUnits=xyUnits,
                        path=path,
                        xMin=xMin, xMax=xMax,
                        xMargin = xMargin,
                        type = self.myType,
                        darkTheme = True)
                        #darkTheme=self.detectionWidget.useDarkStyle)

        tmpLayout.addWidget(self.mainWidget)

        # rewire existing widget into plugin architecture
        #self.mainWidget.closeEvent = self.onClose
        #self._mySetWindowTitle()

        #self.setCentralWidget(self.mainWidget)
        #self.setLayout(tmpLayout)
        self.getVBoxLayout().addLayout(tmpLayout)
        
if __name__ == '__main__':
    path = '/Users/cudmore/Sites/SanPy/data/19114001.abf'
    ba = sanpy.bAnalysis(path)
    ba.spikeDetect()
    print(ba.numSpikes)

    import sys
    app = QtWidgets.QApplication([])
    et = exportTrace(ba=ba)
    et.show()

    path2 = '/Users/cudmore/Sites/SanPy/data/19114000.abf'
    ba2 = sanpy.bAnalysis(path2)
    et.mainWidget.switchFile(ba2)

    sys.exit(app.exec_())
